from typing import List

from .common import EncoderModule, _take, make_n_channel_input
from ..activations import ABN
from ..backbone.wider_resnet import WiderResNet, WiderResNetA2

__all__ = [
    "WiderResnetEncoder",
    "WiderResnet16A2Encoder",
    "WiderResnet16Encoder",
    "WiderResnet20Encoder",
    "WiderResnet38A2Encoder",
    "WiderResnet38Encoder",
    "WiderResnetA2Encoder",
    "WiderResnet20A2Encoder",
]


class WiderResnetEncoder(EncoderModule):
    def __init__(self, structure: List[int], layers: List[int], norm_act=ABN):
        super().__init__([64, 128, 256, 512, 1024, 2048, 4096], [1, 2, 4, 8, 16, 32, 32], layers)

        encoder: WiderResNet = WiderResNet(structure, classes=0, norm_act=norm_act)
        self.layer0 = encoder.mod1
        self.layer1 = encoder.mod2
        self.layer2 = encoder.mod3
        self.layer3 = encoder.mod4
        self.layer4 = encoder.mod5
        self.layer5 = encoder.mod6
        self.layer6 = encoder.mod7

        self.pool2 = encoder.pool2
        self.pool3 = encoder.pool3
        self.pool4 = encoder.pool4
        self.pool5 = encoder.pool5
        self.pool6 = encoder.pool6

    @property
    def encoder_layers(self):
        return [self.layer0, self.layer1, self.layer2, self.layer3, self.layer4, self.layer5, self.layer6]

    def forward(self, input):
        output_features = []

        x = self.layer0(input)
        output_features.append(x)

        x = self.layer1(self.pool2(x))
        output_features.append(x)

        x = self.layer2(self.pool3(x))
        output_features.append(x)

        x = self.layer3(self.pool4(x))
        output_features.append(x)

        x = self.layer4(self.pool5(x))
        output_features.append(x)

        x = self.layer5(self.pool6(x))
        output_features.append(x)

        x = self.layer6(x)
        output_features.append(x)

        # Return only features that were requested
        return _take(output_features, self._layers)

    def change_input_channels(self, input_channels: int, mode="auto"):
        self.layer0.conv1 = make_n_channel_input(self.layer0.conv1, input_channels, mode)
        return self


class WiderResnet16Encoder(WiderResnetEncoder):
    def __init__(self, layers=None):
        if layers is None:
            layers = [2, 3, 4, 5, 6]
        super().__init__(structure=[1, 1, 1, 1, 1, 1], layers=layers)


class WiderResnet20Encoder(WiderResnetEncoder):
    def __init__(self, layers=None):
        if layers is None:
            layers = [2, 3, 4, 5, 6]
        super().__init__(structure=[1, 1, 1, 3, 1, 1], layers=layers)


class WiderResnet38Encoder(WiderResnetEncoder):
    def __init__(self, layers=None):
        if layers is None:
            layers = [2, 3, 4, 5, 6]
        super().__init__(structure=[3, 3, 6, 3, 1, 1], layers=layers)


class WiderResnetA2Encoder(EncoderModule):
    def __init__(self, structure: List[int], layers: List[int], norm_act=ABN):
        super().__init__([64, 128, 256, 512, 1024, 2048, 4096], [1, 2, 4, 8, 16, 32, 32], layers)

        encoder = WiderResNetA2(structure=structure, classes=0, norm_act=norm_act)
        self.layer0 = encoder.mod1
        self.layer1 = encoder.mod2
        self.layer2 = encoder.mod3
        self.layer3 = encoder.mod4
        self.layer4 = encoder.mod5
        self.layer5 = encoder.mod6
        self.layer6 = encoder.mod7

        self.pool2 = encoder.pool2
        self.pool3 = encoder.pool3

    @property
    def encoder_layers(self):
        return [self.layer0, self.layer1, self.layer2, self.layer3, self.layer4, self.layer5, self.layer6]

    def forward(self, input):
        output_features = []

        out = self.layer0(input)
        output_features.append(out)

        out = self.layer1(self.pool2(out))
        output_features.append(out)

        out = self.layer2(self.pool3(out))
        output_features.append(out)

        out = self.layer3(out)
        output_features.append(out)

        out = self.layer4(out)
        output_features.append(out)

        out = self.layer5(out)
        output_features.append(out)

        out = self.layer6(out)
        output_features.append(out)

        # Return only features that were requested
        return _take(output_features, self._layers)

    def change_input_channels(self, input_channels: int, mode="auto"):
        self.layer0.conv1 = make_n_channel_input(self.layer0.conv1, input_channels, mode)
        return self


class WiderResnet16A2Encoder(WiderResnetA2Encoder):
    def __init__(self, layers=None):
        if layers is None:
            layers = [2, 3, 4, 5, 6]
        super().__init__(structure=[1, 1, 1, 1, 1, 1], layers=layers)


class WiderResnet20A2Encoder(WiderResnetA2Encoder):
    def __init__(self, layers=None):
        if layers is None:
            layers = [2, 3, 4, 5, 6]
        super().__init__(structure=[1, 1, 1, 3, 1, 1], layers=layers)


class WiderResnet38A2Encoder(WiderResnetA2Encoder):
    def __init__(self, layers=None):
        if layers is None:
            layers = [2, 3, 4, 5, 6]
        super().__init__(structure=[3, 3, 6, 3, 1, 1], layers=layers)



from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

from ..modules.activations import ABN

__all__ = ["UnetBlock", "UnetCentralBlock", "UnetDecoderBlock"]


class UnetBlock(nn.Module):
    """
    Vanilla U-Net block containing of two convolutions interleaved with batch-norm and RELU
    """

    def __init__(self, in_channels: int, out_channels: int, abn_block=ABN):
        super().__init__()
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1, stride=1, bias=False)
        self.abn1 = abn_block(out_channels)
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1, stride=1, bias=False)
        self.abn2 = abn_block(out_channels)

    def forward(self, x):
        x = self.conv1(x)
        x = self.abn1(x)
        x = self.conv2(x)
        x = self.abn2(x)
        return x


class UnetCentralBlock(nn.Module):
    def __init__(self, in_dec_filters: int, out_filters: int, abn_block=ABN):
        super().__init__()
        self.conv1 = nn.Conv2d(in_dec_filters, out_filters, kernel_size=3, padding=1, stride=2, bias=False)
        self.abn1 = abn_block(out_filters)
        self.conv2 = nn.Conv2d(out_filters, out_filters, kernel_size=3, padding=1, bias=False)
        self.abn2 = abn_block(out_filters)

    def forward(self, x):
        x = self.conv1(x)
        x = self.abn1(x)
        x = self.conv2(x)
        x = self.abn2(x)
        return x


class UnetDecoderBlock(nn.Module):
    """"""

    def __init__(
        self,
        in_dec_filters: int,
        in_enc_filters: int,
        out_filters: int,
        abn_block=ABN,
        dropout_rate=0.0,
        scale_factor=None,
        scale_mode="nearest",
        align_corners=None,
    ):
        super(UnetDecoderBlock, self).__init__()

        self.scale_factor = scale_factor
        self.scale_mode = scale_mode
        self.align_corners = align_corners

        self.conv1 = nn.Conv2d(in_dec_filters + in_enc_filters, out_filters, kernel_size=3, padding=1, bias=False)
        self.abn1 = abn_block(out_filters)

        self.drop = nn.Dropout2d(dropout_rate, inplace=False)

        self.conv2 = nn.Conv2d(out_filters, out_filters, kernel_size=3, padding=1, bias=False)
        self.abn2 = abn_block(out_filters)

    def forward(self, x: torch.Tensor, enc: Optional[torch.Tensor] = None) -> torch.Tensor:
        if self.scale_factor is not None:
            x = F.interpolate(
                x, scale_factor=self.scale_factor, mode=self.scale_mode, align_corners=self.align_corners
            )
        else:
            lat_size = enc.size()[2:]
            x = F.interpolate(x, size=lat_size, mode=self.scale_mode, align_corners=self.align_corners)

        if enc is not None:
            x = torch.cat([x, enc], dim=1)

        x = self.conv1(x)
        x = self.abn1(x)

        x = self.drop(x)

        x = self.conv2(x)
        x = self.abn2(x)

        return x



from typing import Union, Tuple, List, Iterable, Any

__all__ = ["VolumeSlicer", "VolumeMerger"]
import numpy as np
import torch

from pytorch_toolbelt.inference import compute_pyramid_patch_weight_loss


class VolumeSlicer:
    """
    Helper class to slice 3d volume into smaller volumes and merge them back
    """

    def __init__(
        self,
        volume_shape: Tuple[int, int, int],
        voxel_size: Union[int, Tuple[int, int, int]],
        voxel_step: Union[int, Tuple[int, int, int]],
        weight="mean",
    ):
        """

        :param volume_shape: Shape of the source image (D, H, W, C)
        :param voxel_size: Tile size (Scalar or tuple (D, H, W, C)
        :param voxel_step: Step in pixels between voxels (Scalar or tuple (D, H, W))
        :param weight: Fusion algorithm. 'mean' - avergaing
        """
        self.volume_shape = np.array(volume_shape)[:3]

        if isinstance(voxel_size, (tuple, list)):
            if len(voxel_size) != 3:
                raise ValueError()
            self.tile_size = np.array(voxel_size, dtype=int)
        else:
            self.tile_size = np.array([int(voxel_size)] * 3)

        if isinstance(voxel_step, (tuple, list)):
            if len(voxel_step) != 3:
                raise ValueError()
            self.tile_step = np.array(voxel_step, dtype=int)
        else:
            self.tile_step = np.array([int(voxel_step)] * 3)

        self.weight = weight

        if self.tile_step[0] < 1 or self.tile_step[0] > self.tile_size[0]:
            raise ValueError()
        if self.tile_step[1] < 1 or self.tile_step[1] > self.tile_size[1]:
            raise ValueError()
        if self.tile_step[2] < 1 or self.tile_step[2] > self.tile_size[2]:
            raise ValueError()

        overlap = self.tile_size - self.tile_step

        # In case margin is not set, we compute it manually

        self.num_tiles = np.maximum(1, np.ceil((self.volume_shape - overlap) / self.tile_step)).astype(int)
        self.extra_pad = self.tile_step * self.num_tiles - (self.volume_shape - overlap)
        self.pad_before = self.extra_pad // 2
        self.pad_after = self.extra_pad - self.pad_before
        self.orignal_image_roi = (
            slice(self.pad_before[0], self.pad_before[0] + self.volume_shape[0]),
            slice(self.pad_before[1], self.pad_before[1] + self.volume_shape[1]),
            slice(self.pad_before[2], self.pad_before[2] + self.volume_shape[2]),
        )
        self.orignal_mask_roi = (
            slice(None),
            slice(self.pad_before[0], self.pad_before[0] + self.volume_shape[0]),
            slice(self.pad_before[1], self.pad_before[1] + self.volume_shape[1]),
            slice(self.pad_before[2], self.pad_before[2] + self.volume_shape[2]),
        )

        rois = []
        bbox_crops = []

        dim_i = range(0, self.volume_shape[0] + self.extra_pad[0] - self.tile_size[0] + 1, self.tile_step[0])
        dim_j = range(0, self.volume_shape[1] + self.extra_pad[1] - self.tile_size[1] + 1, self.tile_step[1])
        dim_k = range(0, self.volume_shape[2] + self.extra_pad[2] - self.tile_size[2] + 1, self.tile_step[2])

        for i in dim_i:
            for j in dim_j:
                for k in dim_k:
                    roi = (
                        slice(i, i + self.tile_size[0]),
                        slice(j, j + self.tile_size[1]),
                        slice(k, k + self.tile_size[2]),
                    )
                    roi2 = (
                        slice(i - self.pad_before[0], i + self.tile_size[0]),
                        slice(j - self.pad_before[1], j + self.tile_size[1]),
                        slice(k - self.pad_before[2], k + self.tile_size[2]),
                    )
                    rois.append(roi)
                    bbox_crops.append(roi2)

        self.crops = rois
        self.bbox_crops = bbox_crops

    def split(self, volume: np.ndarray, value=0) -> List[np.ndarray]:
        if (volume.shape != self.volume_shape).any():
            raise ValueError(f"Volume shape {volume.shape} is not equal to the expected {self.volume_shape}")

        pad_width = np.stack([self.pad_before, self.pad_after], axis=-1)
        image_pad = np.pad(volume, pad_width, mode="constant", constant_values=value)

        tiles = []
        for roi in self.crops:
            tile = image_pad[roi].copy()
            tiles.append(tile)

        return tiles

    def iter_split(self, volume, value=0) -> Iterable[Tuple[np.ndarray, Any]]:
        if (volume.shape != self.volume_shape).any():
            raise ValueError(f"Volume shape {volume.shape} is not equal to the expected {self.volume_shape}")

        pad_width = np.stack([self.pad_before, self.pad_after], axis=-1)
        image_pad = np.pad(volume, pad_width, mode="constant", constant_values=value)

        for roi in self.crops:
            tile = image_pad[roi].copy()
            yield tile, roi

    @property
    def target_shape(self):
        target_shape = self.volume_shape + self.extra_pad
        return target_shape

    def merge(self, tiles: List[np.ndarray], dtype=np.float32):
        if len(tiles) != len(self.crops):
            raise ValueError

        channels = 1 if len(tiles[0].shape) == 2 else tiles[0].shape[2]
        target_shape = (
            self.image_height + self.margin_bottom + self.margin_top,
            self.image_width + self.margin_right + self.margin_left,
            channels,
        )

        image = np.zeros(target_shape, dtype=np.float64)
        norm_mask = np.zeros(target_shape, dtype=np.float64)

        w = np.dstack([self.weight] * channels)

        for tile, (x, y, tile_width, tile_height) in zip(tiles, self.crops):
            # print(x, y, tile_width, tile_height, image.shape)
            image[y : y + tile_height, x : x + tile_width] += tile * w
            norm_mask[y : y + tile_height, x : x + tile_width] += w

        # print(norm_mask.min(), norm_mask.max())
        norm_mask = np.clip(norm_mask, a_min=np.finfo(norm_mask.dtype).eps, a_max=None)
        normalized = np.divide(image, norm_mask).astype(dtype)
        crop = self.crop_to_orignal_size(normalized)
        return crop

    def crop_to_orignal_size(self, volume):
        return volume[self.orignal_image_roi]

    def _mean(self, volume_size):
        return np.ones(volume_size, dtype=np.float32)

    def _pyramid(self, tile_size):
        w, _, _ = compute_pyramid_patch_weight_loss(tile_size[0], tile_size[1])
        return w


class VolumeMerger:
    """
    Helper class to merge volume segmentations on CPU/GPU.
    """

    def __init__(self, volume_shape, channels: int, weight, device="cpu", dtype=torch.float32):
        """

        :param volume_shape: Shape of the source image
        :param weight: Weighting matrix
        """
        self.channels = channels

        self.weight = torch.from_numpy(np.expand_dims(weight, axis=0)).to(device=device, dtype=dtype)
        self.volume = torch.zeros((channels, *volume_shape), device=device, dtype=dtype)
        self.norm_mask = torch.zeros((1, *volume_shape), device=device, dtype=dtype)

    def accumulate_single(self, tile: torch.Tensor, roi):
        """
        Accumulates single element
        :param sample: Predicted image of shape [C,D,H,W]
        :param crop_coords: Corresponding tile crops w.r.t to original image
        """
        tile = tile.to(device=self.volume.device)
        self.volume[:, roi] += tile * self.weight
        self.norm_mask[:, roi] += self.weight

    def integrate_batch(self, batch: torch.Tensor, rois):
        """
        Accumulates batch of tile predictions
        :param batch: Predicted tiles  of shape [B,C,D,H,W]
        :param crop_coords: Corresponding tile crops w.r.t to original image
        """
        if len(batch) != len(rois):
            raise ValueError("Number of images in batch does not correspond to number of coordinates")

        batch = batch.to(device=self.volume.device)
        for tile, roi in zip(batch, rois):
            weighted_tile = tile * self.weight
            self.volume[:, roi[0], roi[1], roi[2]] += weighted_tile.type_as(self.volume)
            self.norm_mask[:, roi[0], roi[1], roi[2]] += self.weight

    def merge(self) -> torch.Tensor:
        return self.volume / self.norm_mask



from .criterions import *
from .callbacks import *
from .metrics import *
from .opl import *
from .visualization import *
from .utils import *
from .loss_adapter import *


def _register_modules(r):
    from pytorch_toolbelt.modules import encoders as e

    r.add_from_module(e, prefix="tbt.")


def _register_criterions(r):
    from pytorch_toolbelt import losses

    r.add_from_module(losses, prefix="tbt.")


def _register_callbacks(r):
    from pytorch_toolbelt.utils.catalyst import callbacks as c

    r.add_from_module(c, prefix="tbt.")


def register_toolbelt_in_catalyst():
    """
    Register modules, losses & callbacks from pytorch-toolbelt in Catalyst
    """
    from catalyst.registry import MODULE, CRITERION

    MODULE.late_add(_register_modules)
    CRITERION.late_add(_register_criterions)



from functools import partial

import torch
from torch.nn.modules.loss import _Loss

from .functional import focal_loss_with_logits

__all__ = ["BinaryFocalLoss", "FocalLoss"]


class BinaryFocalLoss(_Loss):
    def __init__(
        self,
        alpha=None,
        gamma: float = 2.0,
        ignore_index=None,
        reduction="mean",
        normalized=False,
        reduced_threshold=None,
    ):
        """

        :param alpha: Prior probability of having positive value in target.
        :param gamma: Power factor for dampening weight (focal strenght).
        :param ignore_index: If not None, targets may contain values to be ignored.
        Target values equal to ignore_index will be ignored from loss computation.
        :param reduced: Switch to reduced focal loss. Note, when using this mode you should use `reduction="sum"`.
        :param threshold:
        """
        super().__init__()
        self.ignore_index = ignore_index
        self.focal_loss_fn = partial(
            focal_loss_with_logits,
            alpha=alpha,
            gamma=gamma,
            reduced_threshold=reduced_threshold,
            reduction=reduction,
            normalized=normalized,
            ignore_index=ignore_index,
        )

    def forward(self, label_input, label_target):
        """Compute focal loss for binary classification problem."""
        loss = self.focal_loss_fn(label_input, label_target)
        return loss


class FocalLoss(_Loss):
    def __init__(
        self, alpha=None, gamma=2, ignore_index=None, reduction="mean", normalized=False, reduced_threshold=None
    ):
        """
        Focal loss for multi-class problem.

        :param alpha:
        :param gamma:
        :param ignore_index: If not None, targets with given index are ignored
        :param reduced_threshold: A threshold factor for computing reduced focal loss
        """
        super().__init__()
        self.ignore_index = ignore_index
        self.focal_loss_fn = partial(
            focal_loss_with_logits,
            alpha=alpha,
            gamma=gamma,
            reduced_threshold=reduced_threshold,
            reduction=reduction,
            normalized=normalized,
        )

    def forward(self, label_input, label_target):
        num_classes = label_input.size(1)
        loss = 0

        # Filter anchors with -1 label from loss computation
        if self.ignore_index is not None:
            not_ignored = label_target != self.ignore_index

        for cls in range(num_classes):
            cls_label_target = (label_target == cls).long()
            cls_label_input = label_input[:, cls, ...]

            if self.ignore_index is not None:
                cls_label_target = cls_label_target[not_ignored]
                cls_label_input = cls_label_input[not_ignored]

            loss += self.focal_loss_fn(cls_label_input, cls_label_target)
        return loss



from typing import Type, Dict, Any

from torch.optim.optimizer import Optimizer
from torch import nn
import torch.optim
import copy


__all__ = ["get_optimizer", "get_optimizer_cls", "scale_learning_rate_for_ddp"]


def get_optimizer_cls(optimizer_name) -> Type[Optimizer]:
    _OPTIMIZERS_REGISTRY = {
        "sgd": torch.optim.SGD,
        "asgd": torch.optim.ASGD,
        "adam": torch.optim.Adam,
        "adamw": torch.optim.AdamW,
        "adamax": torch.optim.Adamax,
        "adagrad": torch.optim.Adagrad,
        "adadelta": torch.optim.Adadelta,
        "rmsprop": torch.optim.RMSprop,
        "rprop": torch.optim.Rprop,
    }

    optimizer_name = optimizer_name.lower()
    if optimizer_name in _OPTIMIZERS_REGISTRY:
        return _OPTIMIZERS_REGISTRY[optimizer_name]

    _APEX_OPTIMIZERS = ["fused_adam", "fused_sgd", "fused_lamb"]
    if optimizer_name in _APEX_OPTIMIZERS:
        try:
            from apex.optimizers import FusedAdam, FusedSGD, FusedLAMB

            _APEX_REGISTRY = {"fused_adam": FusedAdam, "fused_sgd": FusedSGD, "fused_lamb": FusedLAMB}
            if optimizer_name in _OPTIMIZERS_REGISTRY:
                return _APEX_REGISTRY[optimizer_name]
        except ImportError:
            raise ValueError(
                f"Requested optimizer {optimizer_name} is not available since NVIDIA/apex is not installed"
            )

    try:
        import torch_optimizer as to

        return to.get(optimizer_name)
    except ImportError:
        raise ValueError(
            f"Requested optimizer {optimizer_name} is not available since torch-optimizer is not installed. "
            f"Please install it from https://github.com/jettify/pytorch-optimizer"
        )
    except ValueError:
        raise


def scale_learning_rate_for_ddp(optimizer_params: Dict[str, Any]) -> Dict[str, Any]:
    """
    Scale learning rate with respect to world size for Distributed Data Parallel training.
    Efficient learning rate is WORLD_SIZE * learning rate from config
    For non-distributed runs WORLD_SIZE is 1
    """
    if torch.distributed.is_available() and torch.distributed.is_initialized():
        optimizer_params = copy.deepcopy(optimizer_params)
        optimizer_params["lr"] = optimizer_params["lr"] * torch.distributed.get_world_size()

    return optimizer_params


def get_optimizer(
    model: nn.Module,
    optimizer_name: str,
    optimizer_params: Dict[str, Any],
    apply_weight_decay_to_bias: bool = True,
) -> Optimizer:
    """
    Construct an Optimizer for given model
    Args:
        model: Model to optimize. Only parameters that require_grad will be used
        optimizer_name: Name of the optimizer (case-insensitive). Supports native pytorch optimizers, apex and
            optimizers from pytorch-optimizers package.
        optimizer_params: Dict of optimizer params (lr, weight_decay, eps, etc)
        apply_weight_decay_to_bias: Whether to apply weight decay on bias parameters. Default is True
    Returns:
        Optimizer instance
    """
    # Optimizer parameter groups
    default_pg, biases_pg = [], []

    for k, v in model.named_parameters():
        if v.requires_grad:
            if str.endswith(k, ".bias"):
                biases_pg.append(v)  # biases
            else:
                default_pg.append(v)  # all else

    if apply_weight_decay_to_bias:
        parameters = default_pg + biases_pg
    else:
        parameters = default_pg

    optimizer_cls = get_optimizer_cls(optimizer_name)

    optimizer: Optimizer = optimizer_cls(
        parameters,
        **optimizer_params,
    )

    if not apply_weight_decay_to_bias:
        optimizer.add_param_group({"params": biases_pg, "weight_decay": 0.0})

    return optimizer



from .common import GenericTimmEncoder, make_n_channel_input_std_conv


__all__ = [
    "NFRegNetB0Encoder",
    "NFRegNetB1Encoder",
    "NFRegNetB2Encoder",
    "NFRegNetB3Encoder",
    "NFRegNetB4Encoder",
    "NFRegNetB5Encoder",
]


class NFRegNetB0Encoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nf_regnet_b0(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv = make_n_channel_input_std_conv(self.encoder.stem_conv, input_channels, mode, **kwargs)
        return self


class NFRegNetB1Encoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nf_regnet_b1(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv = make_n_channel_input_std_conv(self.encoder.stem_conv, input_channels, mode, **kwargs)
        return self


class NFRegNetB2Encoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nf_regnet_b2(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv = make_n_channel_input_std_conv(self.encoder.stem_conv, input_channels, mode, **kwargs)
        return self


class NFRegNetB3Encoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nf_regnet_b3(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv = make_n_channel_input_std_conv(self.encoder.stem_conv, input_channels, mode, **kwargs)
        return self


class NFRegNetB4Encoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nf_regnet_b4(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv = make_n_channel_input_std_conv(self.encoder.stem_conv, input_channels, mode, **kwargs)
        return self


class NFRegNetB5Encoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nf_regnet_b5(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv = make_n_channel_input_std_conv(self.encoder.stem_conv, input_channels, mode, **kwargs)
        return self



from collections import OrderedDict
from typing import List

import torch
import torch.nn.functional as F
from torch import nn

from .common import EncoderModule, make_n_channel_input, _take

__all__ = ["HRNetV2Encoder18", "HRNetV2Encoder34", "HRNetV2Encoder48"]


HRNETV2_BN_MOMENTUM = 0.1


def hrnet_conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False)


class HRNetBasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(HRNetBasicBlock, self).__init__()
        self.conv1 = hrnet_conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes, momentum=HRNETV2_BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = hrnet_conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes, momentum=HRNETV2_BN_MOMENTUM)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class HRNetBottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(HRNetBottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes, momentum=HRNETV2_BN_MOMENTUM)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes, momentum=HRNETV2_BN_MOMENTUM)
        self.conv3 = nn.Conv2d(planes, planes * self.expansion, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * self.expansion, momentum=HRNETV2_BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class HighResolutionModule(nn.Module):
    def __init__(
        self, num_branches, blocks, num_blocks, num_inchannels, num_channels, fuse_method, multi_scale_output=True
    ):
        super(HighResolutionModule, self).__init__()
        self._check_branches(num_branches, blocks, num_blocks, num_inchannels, num_channels)

        self.num_inchannels = num_inchannels
        self.fuse_method = fuse_method
        self.num_branches = num_branches

        self.multi_scale_output = multi_scale_output

        self.branches = self._make_branches(num_branches, blocks, num_blocks, num_channels)
        self.fuse_layers = self._make_fuse_layers()
        self.relu = nn.ReLU(inplace=True)

    def _check_branches(self, num_branches, blocks, num_blocks, num_inchannels, num_channels):
        if num_branches != len(num_blocks):
            error_msg = "NUM_BRANCHES({}) <> NUM_BLOCKS({})".format(num_branches, len(num_blocks))
            raise ValueError(error_msg)

        if num_branches != len(num_channels):
            error_msg = "NUM_BRANCHES({}) <> NUM_CHANNELS({})".format(num_branches, len(num_channels))
            raise ValueError(error_msg)

        if num_branches != len(num_inchannels):
            error_msg = "NUM_BRANCHES({}) <> NUM_INCHANNELS({})".format(num_branches, len(num_inchannels))
            raise ValueError(error_msg)

    def _make_one_branch(self, branch_index, block, num_blocks, num_channels, stride=1):
        downsample = None
        if stride != 1 or self.num_inchannels[branch_index] != num_channels[branch_index] * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(
                    self.num_inchannels[branch_index],
                    num_channels[branch_index] * block.expansion,
                    kernel_size=1,
                    stride=stride,
                    bias=False,
                ),
                nn.BatchNorm2d(num_channels[branch_index] * block.expansion, momentum=HRNETV2_BN_MOMENTUM),
            )

        layers = []
        layers.append(block(self.num_inchannels[branch_index], num_channels[branch_index], stride, downsample))
        self.num_inchannels[branch_index] = num_channels[branch_index] * block.expansion
        for i in range(1, num_blocks[branch_index]):
            layers.append(block(self.num_inchannels[branch_index], num_channels[branch_index]))

        return nn.Sequential(*layers)

    def _make_branches(self, num_branches, block, num_blocks, num_channels):
        branches = []

        for i in range(num_branches):
            branches.append(self._make_one_branch(i, block, num_blocks, num_channels))

        return nn.ModuleList(branches)

    def _make_fuse_layers(self):
        if self.num_branches == 1:
            return None

        num_branches = self.num_branches
        num_inchannels = self.num_inchannels
        fuse_layers = []
        for i in range(num_branches if self.multi_scale_output else 1):
            fuse_layer = []
            for j in range(num_branches):
                if j > i:
                    fuse_layer.append(
                        nn.Sequential(
                            nn.Conv2d(num_inchannels[j], num_inchannels[i], 1, 1, 0, bias=False),
                            nn.BatchNorm2d(num_inchannels[i], momentum=HRNETV2_BN_MOMENTUM),
                        )
                    )
                elif j == i:
                    fuse_layer.append(None)
                else:
                    conv3x3s = []
                    for k in range(i - j):
                        if k == i - j - 1:
                            num_outchannels_conv3x3 = num_inchannels[i]
                            conv3x3s.append(
                                nn.Sequential(
                                    nn.Conv2d(num_inchannels[j], num_outchannels_conv3x3, 3, 2, 1, bias=False),
                                    nn.BatchNorm2d(num_outchannels_conv3x3, momentum=HRNETV2_BN_MOMENTUM),
                                )
                            )
                        else:
                            num_outchannels_conv3x3 = num_inchannels[j]
                            conv3x3s.append(
                                nn.Sequential(
                                    nn.Conv2d(num_inchannels[j], num_outchannels_conv3x3, 3, 2, 1, bias=False),
                                    nn.BatchNorm2d(num_outchannels_conv3x3, momentum=HRNETV2_BN_MOMENTUM),
                                    nn.ReLU(inplace=True),
                                )
                            )
                    fuse_layer.append(nn.Sequential(*conv3x3s))
            fuse_layers.append(nn.ModuleList(fuse_layer))

        return nn.ModuleList(fuse_layers)

    def get_num_inchannels(self):
        return self.num_inchannels

    def forward(self, x):
        if self.num_branches == 1:
            return [self.branches[0](x[0])]

        for i in range(self.num_branches):
            x[i] = self.branches[i](x[i])

        x_fuse = []
        for i in range(len(self.fuse_layers)):
            y = x[0] if i == 0 else self.fuse_layers[i][0](x[0])
            for j in range(1, self.num_branches):
                if i == j:
                    y = y + x[j]
                elif j > i:
                    width_output = x[i].shape[-1]
                    height_output = x[i].shape[-2]
                    y = y + F.interpolate(
                        self.fuse_layers[i][j](x[j]),
                        size=(height_output, width_output),
                        mode="bilinear",
                        align_corners=False,
                    )
                else:
                    y = y + self.fuse_layers[i][j](x[j])
            x_fuse.append(self.relu(y))

        return x_fuse


class HRNetEncoderBase(EncoderModule):
    def __init__(self, input_channels=3, width=48, layers: List[int] = None):
        if layers is None:
            layers = [1, 2, 3, 4]

        channels = [64, width, width * 2, width * 4, width * 8]

        strides = [4, 4, 8, 16, 32]

        super().__init__(channels=channels, strides=strides, layers=layers)

        blocks_dict = {"BASIC": HRNetBasicBlock, "BOTTLENECK": HRNetBottleneck}

        extra = {
            "STAGE2": {
                "NUM_MODULES": 1,
                "NUM_BRANCHES": 2,
                "BLOCK": "BASIC",
                "NUM_BLOCKS": (4, 4),
                "NUM_CHANNELS": (width, width * 2),
                "FUSE_METHOD": "SUM",
            },
            "STAGE3": {
                "NUM_MODULES": 4,
                "NUM_BRANCHES": 3,
                "BLOCK": "BASIC",
                "NUM_BLOCKS": (4, 4, 4),
                "NUM_CHANNELS": (width, width * 2, width * 4),
                "FUSE_METHOD": "SUM",
            },
            "STAGE4": {
                "NUM_MODULES": 3,
                "NUM_BRANCHES": 4,
                "BLOCK": "BASIC",
                "NUM_BLOCKS": (4, 4, 4, 4),
                "NUM_CHANNELS": (width, width * 2, width * 4, width * 8),
                "FUSE_METHOD": "SUM",
            },
            "FINAL_CONV_KERNEL": 1,
        }

        # stem net
        self.layer0 = nn.Sequential(
            OrderedDict(
                [
                    ("conv1", nn.Conv2d(input_channels, 64, kernel_size=3, stride=2, padding=1, bias=False)),
                    ("bn1", nn.BatchNorm2d(64, momentum=HRNETV2_BN_MOMENTUM)),
                    ("relu", nn.ReLU(inplace=True)),
                    ("conv2", nn.Conv2d(64, 64, kernel_size=3, stride=2, padding=1, bias=False)),
                    ("bn2", nn.BatchNorm2d(64, momentum=HRNETV2_BN_MOMENTUM)),
                    ("relu2", nn.ReLU(inplace=True)),
                ]
            )
        )

        self.layer1 = self._make_layer(HRNetBottleneck, 64, 64, 4)

        self.stage2_cfg = extra["STAGE2"]
        num_channels = self.stage2_cfg["NUM_CHANNELS"]
        block = blocks_dict[self.stage2_cfg["BLOCK"]]
        num_channels = [num_channels[i] * block.expansion for i in range(len(num_channels))]
        self.transition1 = self._make_transition_layer([256], num_channels)
        self.stage2, pre_stage_channels = self._make_stage(self.stage2_cfg, num_channels)

        self.stage3_cfg = extra["STAGE3"]
        num_channels = self.stage3_cfg["NUM_CHANNELS"]
        block = blocks_dict[self.stage3_cfg["BLOCK"]]
        num_channels = [num_channels[i] * block.expansion for i in range(len(num_channels))]
        self.transition2 = self._make_transition_layer(pre_stage_channels, num_channels)
        self.stage3, pre_stage_channels = self._make_stage(self.stage3_cfg, num_channels)

        self.stage4_cfg = extra["STAGE4"]
        num_channels = self.stage4_cfg["NUM_CHANNELS"]
        block = blocks_dict[self.stage4_cfg["BLOCK"]]
        num_channels = [num_channels[i] * block.expansion for i in range(len(num_channels))]
        self.transition3 = self._make_transition_layer(pre_stage_channels, num_channels)
        self.stage4, pre_stage_channels = self._make_stage(self.stage4_cfg, num_channels, multi_scale_output=True)

    def _make_transition_layer(self, num_channels_pre_layer, num_channels_cur_layer):
        num_branches_cur = len(num_channels_cur_layer)
        num_branches_pre = len(num_channels_pre_layer)

        transition_layers = []
        for i in range(num_branches_cur):
            if i < num_branches_pre:
                if num_channels_cur_layer[i] != num_channels_pre_layer[i]:
                    transition_layers.append(
                        nn.Sequential(
                            nn.Conv2d(num_channels_pre_layer[i], num_channels_cur_layer[i], 3, 1, 1, bias=False),
                            nn.BatchNorm2d(num_channels_cur_layer[i], momentum=HRNETV2_BN_MOMENTUM),
                            nn.ReLU(inplace=True),
                        )
                    )
                else:
                    transition_layers.append(None)
            else:
                conv3x3s = []
                for j in range(i + 1 - num_branches_pre):
                    inchannels = num_channels_pre_layer[-1]
                    outchannels = num_channels_cur_layer[i] if j == i - num_branches_pre else inchannels
                    conv3x3s.append(
                        nn.Sequential(
                            nn.Conv2d(inchannels, outchannels, 3, 2, 1, bias=False),
                            nn.BatchNorm2d(outchannels, momentum=HRNETV2_BN_MOMENTUM),
                            nn.ReLU(inplace=True),
                        )
                    )
                transition_layers.append(nn.Sequential(*conv3x3s))

        return nn.ModuleList(transition_layers)

    def _make_layer(self, block, inplanes, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(inplanes, planes * block.expansion, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion, momentum=HRNETV2_BN_MOMENTUM),
            )

        layers = []
        layers.append(block(inplanes, planes, stride, downsample))
        inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(inplanes, planes))

        return nn.Sequential(*layers)

    def _make_stage(self, layer_config, num_inchannels, multi_scale_output=True):

        blocks_dict = {"BASIC": HRNetBasicBlock, "BOTTLENECK": HRNetBottleneck}

        num_modules = layer_config["NUM_MODULES"]
        num_branches = layer_config["NUM_BRANCHES"]
        num_blocks = layer_config["NUM_BLOCKS"]
        num_channels = layer_config["NUM_CHANNELS"]
        block = blocks_dict[layer_config["BLOCK"]]
        fuse_method = layer_config["FUSE_METHOD"]

        modules = []
        for i in range(num_modules):
            # multi_scale_output is only used last module
            if not multi_scale_output and i == num_modules - 1:
                reset_multi_scale_output = False
            else:
                reset_multi_scale_output = True
            modules.append(
                HighResolutionModule(
                    num_branches,
                    block,
                    num_blocks,
                    num_inchannels,
                    num_channels,
                    fuse_method,
                    reset_multi_scale_output,
                )
            )
            num_inchannels = modules[-1].get_num_inchannels()

        return nn.Sequential(*modules), num_inchannels

    def forward(self, x):
        layer0 = self.layer0(x)
        x = self.layer1(layer0)

        x_list = []
        for i in range(self.stage2_cfg["NUM_BRANCHES"]):
            if self.transition1[i] is not None:
                x_list.append(self.transition1[i](x))
            else:
                x_list.append(x)
        y_list = self.stage2(x_list)

        x_list = []
        for i in range(self.stage3_cfg["NUM_BRANCHES"]):
            if self.transition2[i] is not None:
                x_list.append(self.transition2[i](y_list[-1]))
            else:
                x_list.append(y_list[i])
        y_list = self.stage3(x_list)

        x_list = []
        for i in range(self.stage4_cfg["NUM_BRANCHES"]):
            if self.transition3[i] is not None:
                x_list.append(self.transition3[i](y_list[-1]))
            else:
                x_list.append(y_list[i])
        y_list = self.stage4(x_list)

        outputs = _take([layer0] + y_list, self._layers)
        return outputs

    def change_input_channels(self, input_channels: int, mode="auto"):
        self.layer0.conv1 = make_n_channel_input(self.layer0.conv1, input_channels, mode)
        return self


class HRNetV2Encoder18(HRNetEncoderBase):
    def __init__(self, pretrained=None, layers=None):
        super().__init__(width=18, layers=layers)


class HRNetV2Encoder34(HRNetEncoderBase):
    def __init__(self, pretrained=None, layers=None):
        super().__init__(width=34, layers=layers)


class HRNetV2Encoder48(HRNetEncoderBase):
    def __init__(self, pretrained=None, layers=None):
        super().__init__(width=48, layers=layers)



from functools import partial
from typing import Optional, List, Callable

import albumentations as A
import cv2
import numpy as np
from skimage.measure import block_reduce
from torch.utils.data import Dataset

from .common import (
    read_image_rgb,
    INPUT_IMAGE_KEY,
    INPUT_IMAGE_ID_KEY,
    INPUT_INDEX_KEY,
    TARGET_MASK_WEIGHT_KEY,
    TARGET_MASK_KEY,
    name_for_stride,
)
from ..utils import fs, image_to_tensor

__all__ = ["mask_to_bce_target", "mask_to_ce_target", "read_binary_mask", "SegmentationDataset", "compute_weight_mask"]


def mask_to_bce_target(mask):
    return image_to_tensor(mask, dummy_channels_dim=True).float()


def mask_to_ce_target(mask):
    return image_to_tensor(mask, dummy_channels_dim=False).long()


def compute_weight_mask(mask: np.ndarray, edge_weight=4) -> np.ndarray:
    from skimage.morphology import binary_dilation, binary_erosion

    binary_mask = mask > 0
    weight_mask = np.ones(mask.shape[:2]).astype(np.float32)

    if binary_mask.any():
        dilated = binary_dilation(binary_mask, structure=np.ones((5, 5), dtype=np.bool))
        eroded = binary_erosion(binary_mask, structure=np.ones((5, 5), dtype=np.bool))

        a = dilated & ~binary_mask
        b = binary_mask & ~eroded

        weight_mask = (a | b).astype(np.float32) * edge_weight + 1
        weight_mask = cv2.GaussianBlur(weight_mask, ksize=(5, 5), sigmaX=5)
    return weight_mask


def _block_reduce_dominant_label(x: np.ndarray, axis):
    try:
        # minlength is +1 to num classes because we must account for IGNORE_LABEL
        minlength = np.max(x) + 1
        bincount_fn = partial(np.bincount, minlength=minlength)
        counts = np.apply_along_axis(bincount_fn, -1, x.reshape((x.shape[0], x.shape[1], -1)))
        reduced = np.argmax(counts, axis=-1)
        return reduced
    except Exception as e:
        print(e)
        print("shape", x.shape, "axis", axis)


def read_binary_mask(mask_fname: str) -> np.ndarray:
    """
    Read image as binary mask, all non-zero values are treated as positive labels and converted to 1
    Args:
        mask_fname: Image with mask

    Returns:
        Numpy array with {0,1} values
    """
    mask = cv2.imread(mask_fname, cv2.IMREAD_GRAYSCALE)
    if mask is None:
        raise FileNotFoundError(f"Cannot find {mask_fname}")

    cv2.threshold(mask, thresh=0, maxval=1, type=cv2.THRESH_BINARY, dst=mask)
    return mask


class SegmentationDataset(Dataset):
    """
    Dataset class suitable for segmentation tasks
    """

    def __init__(
        self,
        image_filenames: List[str],
        mask_filenames: Optional[List[str]],
        transform: A.Compose,
        read_image_fn: Callable = read_image_rgb,
        read_mask_fn: Callable = cv2.imread,
        need_weight_mask=False,
        need_supervision_masks=False,
        make_mask_target_fn: Callable = mask_to_ce_target,
        image_ids: Optional[List[str]] = None,
    ):
        if mask_filenames is not None and len(image_filenames) != len(mask_filenames):
            raise ValueError("Number of images does not corresponds to number of targets")

        if image_ids is None:
            self.image_ids = [fs.id_from_fname(fname) for fname in image_filenames]
        else:
            self.image_ids = image_ids

        self.need_weight_mask = need_weight_mask
        self.need_supervision_masks = need_supervision_masks

        self.images = image_filenames
        self.masks = mask_filenames
        self.read_image = read_image_fn
        self.read_mask = read_mask_fn

        self.transform = transform
        self.make_target = make_mask_target_fn

    def __len__(self):
        return len(self.images)

    def __getitem__(self, index):
        image = self.read_image(self.images[index])
        data = {"image": image}
        if self.masks is not None:
            data["mask"] = self.read_mask(self.masks[index])

        data = self.transform(**data)

        image = data["image"]
        sample = {
            INPUT_INDEX_KEY: index,
            INPUT_IMAGE_ID_KEY: self.image_ids[index],
            INPUT_IMAGE_KEY: image_to_tensor(image),
        }

        if self.masks is not None:
            mask = data["mask"]
            sample[TARGET_MASK_KEY] = self.make_target(mask)
            if self.need_weight_mask:
                sample[TARGET_MASK_WEIGHT_KEY] = image_to_tensor(compute_weight_mask(mask)).float()

            if self.need_supervision_masks:
                for i in range(1, 6):
                    stride = 2**i
                    mask = block_reduce(mask, (2, 2), partial(_block_reduce_dominant_label))
                    sample[name_for_stride(TARGET_MASK_KEY, stride)] = self.make_target(mask)

        return sample



"""Common functions to marshal data to/from PyTorch

"""
import collections
from typing import Optional, Sequence, Union, Dict, List, Any, Iterable

import numpy as np
import torch
from torch import nn, Tensor
from .support import pytorch_toolbelt_deprecated

__all__ = [
    "argmax_over_dim_0",
    "argmax_over_dim_1",
    "argmax_over_dim_2",
    "argmax_over_dim_3",
    "count_parameters",
    "image_to_tensor",
    "int_to_string_human_friendly",
    "logit",
    "mask_from_tensor",
    "maybe_cuda",
    "resize_as",
    "resize_like",
    "rgb_image_from_tensor",
    "sigmoid_with_threshold",
    "softmax_over_dim_0",
    "softmax_over_dim_1",
    "softmax_over_dim_2",
    "softmax_over_dim_3",
    "tensor_from_mask_image",
    "tensor_from_rgb_image",
    "to_numpy",
    "to_tensor",
    "transfer_weights",
    "move_to_device_non_blocking",
    "describe_outputs",
]


def softmax_over_dim_0(x: Tensor) -> Tensor:
    return torch.softmax(x, dim=0)


def softmax_over_dim_1(x: Tensor) -> Tensor:
    return torch.softmax(x, dim=1)


def softmax_over_dim_2(x: Tensor) -> Tensor:
    return torch.softmax(x, dim=2)


def softmax_over_dim_3(x: Tensor) -> Tensor:
    return torch.softmax(x, dim=3)


def argmax_over_dim_0(x: Tensor) -> Tensor:
    return torch.argmax(x, dim=0)


def argmax_over_dim_1(x: Tensor) -> Tensor:
    return torch.argmax(x, dim=1)


def argmax_over_dim_2(x: Tensor) -> Tensor:
    return torch.argmax(x, dim=2)


def argmax_over_dim_3(x: Tensor) -> Tensor:
    return torch.argmax(x, dim=3)


def sigmoid_with_threshold(x: Tensor, threshold):
    return x.float().sigmoid().gt(threshold)


def logit(x: torch.Tensor, eps=1e-5) -> torch.Tensor:
    """
    Compute inverse of sigmoid of the input.
    Note: This function has not been tested for numerical stability.
    :param x:
    :param eps:
    :return:
    """
    x = torch.clamp(x, eps, 1.0 - eps)
    return torch.log(x / (1.0 - x))


def count_parameters(
    model: nn.Module, keys: Optional[Sequence[str]] = None, human_friendly: bool = False
) -> Dict[str, int]:
    """
    Count number of total and trainable parameters of a model
    :param model: A model
    :param keys: Optional list of top-level blocks
    :param human_friendly: If True, outputs human-friendly number of paramters: 13.3M, 124K
    :return: Tuple (total, trainable)
    """
    if keys is None:
        keys = ["encoder", "decoder", "logits", "head", "final"]
    total = int(sum(p.numel() for p in model.parameters()))
    trainable = int(sum(p.numel() for p in model.parameters() if p.requires_grad))
    parameters = {"total": total, "trainable": trainable}

    for key in keys:
        try:
            if hasattr(model, key) and model.__getattr__(key) is not None:
                parameters[key] = int(sum(p.numel() for p in model.__getattr__(key).parameters()))
        except AttributeError:
            pass

    if human_friendly:
        for key in parameters.keys():
            parameters[key] = int_to_string_human_friendly(parameters[key])
    return parameters


def int_to_string_human_friendly(value: int) -> str:
    if value < 1000:
        return str(value)
    if value < 1000000:
        return f"{value / 1000.:.2f}K"
    if value < 10000000:
        return f"{value / 1000000.:.2f}M"
    if value < 100000000:
        return f"{value / 1000000.:.1f}M"
    if value < 1000000000:
        return f"{value / 1000000.:.1f}M"
    return f"{value / 1000000000.:.2f}B"


def to_numpy(x: Union[torch.Tensor, np.ndarray, Any, None]) -> Union[np.ndarray, None]:
    """
    Convert whatever to numpy array. None value returned as is.

    Args:
        :param x: List, tuple, PyTorch tensor or numpy array

    Returns:
        :return: Numpy array
    """
    if x is None:
        return None
    elif torch.is_tensor(x):
        return x.data.cpu().numpy()
    elif isinstance(x, np.ndarray):
        return x
    elif isinstance(x, (Iterable, int, float)):
        return np.array(x)
    else:
        raise ValueError("Unsupported type")


def to_tensor(x, dtype=None) -> torch.Tensor:
    if isinstance(x, torch.Tensor):
        if dtype is not None:
            x = x.type(dtype)
        return x
    if isinstance(x, np.ndarray) and x.dtype.kind not in {"O", "M", "U", "S"}:
        x = torch.from_numpy(x)
        if dtype is not None:
            x = x.type(dtype)
        return x
    if isinstance(x, (list, tuple)):
        x = np.ndarray(x)
        x = torch.from_numpy(x)
        if dtype is not None:
            x = x.type(dtype)
        return x

    raise ValueError("Unsupported input type" + str(type(x)))


def image_to_tensor(image: np.ndarray, dummy_channels_dim=True) -> torch.Tensor:
    """
    Convert numpy image (RGB, BGR, Grayscale, SAR, Mask image, etc.) to tensor

    Args:
        image: A numpy array of [H,W,C] shape
        dummy_channels_dim: If True, and image has [H,W] shape adds dummy channel, so that
            output tensor has shape [1, H, W]

    See also:
        rgb_image_from_tensor - To convert tensor image back to RGB with denormalization
        mask_from_tensor

    Returns:
        Torch tensor of [C,H,W] or [H,W] shape (dummy_channels_dim=False).
    """
    if len(image.shape) not in {2, 3}:
        raise ValueError(f"Image must have shape [H,W] or [H,W,C]. Got image with shape {image.shape}")

    if len(image.shape) == 2:
        if dummy_channels_dim:
            image = np.expand_dims(image, 0)
    else:
        # HWC -> CHW
        image = np.moveaxis(image, -1, 0)
    image = np.ascontiguousarray(image)
    image = torch.from_numpy(image)
    return image


@pytorch_toolbelt_deprecated("This function is deprecated, please use image_to_tensor instead")
def tensor_from_rgb_image(image: np.ndarray) -> torch.Tensor:
    return image_to_tensor(image)


@pytorch_toolbelt_deprecated("This function is deprecated, please use image_to_tensor instead")
def tensor_from_mask_image(mask: np.ndarray) -> torch.Tensor:
    return image_to_tensor(mask)


def rgb_image_from_tensor(
    image: torch.Tensor,
    mean=(0.485, 0.456, 0.406),
    std=(0.229, 0.224, 0.225),
    min_pixel_value=0.0,
    max_pixel_value=255.0,
    dtype=np.uint8,
) -> np.ndarray:
    """
    Convert numpy image (RGB, BGR, Grayscale, SAR, Mask image, etc.) to tensor

    Args:
        image: A torch tensor of [C,H,W] shape
    """
    image = np.moveaxis(to_numpy(image), 0, -1)
    mean = to_numpy(mean)
    std = to_numpy(std)
    rgb_image = max_pixel_value * (image * std + mean)
    rgb_image = np.clip(rgb_image, a_min=min_pixel_value, a_max=max_pixel_value)
    return rgb_image.astype(dtype)


def mask_from_tensor(mask: torch.Tensor, squeeze_single_channel: bool = False, dtype=None) -> np.ndarray:
    mask_np = np.moveaxis(to_numpy(mask), 0, -1)
    if squeeze_single_channel and mask_np.shape[-1] == 1:
        mask_np = np.squeeze(mask_np, -1)

    if dtype is not None:
        mask_np = mask_np.astype(dtype)
    return mask_np


def maybe_cuda(x: Union[torch.Tensor, nn.Module]) -> Union[torch.Tensor, nn.Module]:
    """
    Move input Tensor or Module to CUDA device if CUDA is available.
    :param x:
    :return:
    """
    if torch.cuda.is_available():
        return x.cuda()
    return x


def transfer_weights(model: nn.Module, model_state_dict: collections.OrderedDict):
    """
    Copy weights from state dict to model, skipping layers that are incompatible.
    This method is helpful if you are doing some model surgery and want to load
    part of the model weights into different model.
    :param model: Model to load weights into
    :param model_state_dict: Model state dict to load weights from
    :return: None
    """
    for name, value in model_state_dict.items():
        try:
            model.load_state_dict(collections.OrderedDict([(name, value)]), strict=False)
        except Exception as e:
            print(e)


def resize_like(x: Tensor, target: Tensor, mode: str = "bilinear", align_corners: bool = True) -> Tensor:
    """
    Resize input tensor to have the same spatial dimensions as target.

    Args:
        x: Input tensor of [B,C,H,W]
        target: [Bt,Ct,Ht,Wt]
        mode:
        align_corners:

    Returns:
        Resized tensor [B,C,Ht,Wt]
    """
    return torch.nn.functional.interpolate(x, target.size()[2:], mode=mode, align_corners=align_corners)


def move_to_device_non_blocking(x: Tensor, device: torch.device) -> Tensor:
    if x.device != device:
        x = x.to(device=device, non_blocking=True)
    return x


resize_as = resize_like


def describe_outputs(outputs: Union[Tensor, Dict[str, Tensor], Iterable[Tensor]]) -> Union[List[Dict], Dict[str, Any]]:
    """
    Describe outputs and return shape, mean & std for each tensor in list or dict (Supports nested tensors)

    Args:
        outputs: Input (Usually model outputs)
    Returns:
        Same structure but each item represents tensor shape, mean & std
    """
    if torch.is_tensor(outputs):
        desc = dict(size=tuple(outputs.size()), mean=outputs.mean().item(), std=outputs.std().item())
    elif isinstance(outputs, collections.Mapping):
        desc = {}
        for key, value in outputs.items():
            desc[key] = describe_outputs(value)
    elif isinstance(outputs, collections.Iterable):
        desc = []
        for index, output in enumerate(outputs):
            desc.append(describe_outputs(output))
    else:
        raise NotImplementedError
    return desc



"""Implementation of different pooling modules

"""
from typing import Union, Dict

import torch
import torch.nn.functional as F
from torch.nn.modules.module import _IncompatibleKeys
from torch import Tensor, nn

__all__ = [
    "GWAP",
    "GlobalAvgPool2d",
    "GlobalKMaxPool2d",
    "GlobalMaxPool2d",
    "GlobalRankPooling",
    "GeneralizedMeanPooling2d",
    "GlobalWeightedAvgPool2d",
    "MILCustomPoolingModule",
    "RMSPool",
]


class GlobalAvgPool2d(nn.Module):
    def __init__(self, flatten=False):
        """Global average pooling over the input's spatial dimensions"""
        super().__init__()
        self.flatten = flatten

    def forward(self, x):  # skipcq: PYL-W0221
        x = F.adaptive_avg_pool2d(x, output_size=1)
        if self.flatten:
            x = x.view(x.size(0), x.size(1))
        return x


class GlobalMaxPool2d(nn.Module):
    def __init__(self, flatten=False):
        """Global average pooling over the input's spatial dimensions"""
        super().__init__()
        self.flatten = flatten

    def forward(self, x):  # skipcq: PYL-W0221
        x = F.adaptive_max_pool2d(x, output_size=1)
        if self.flatten:
            x = x.view(x.size(0), x.size(1))
        return x


class GlobalKMaxPool2d(nn.Module):
    """
    K-max global pooling block

    https://arxiv.org/abs/1911.07344
    """

    def __init__(self, k=4, trainable=True, flatten=False):
        """Global average pooling over the input's spatial dimensions"""
        super().__init__()
        self.k = k
        self.flatten = flatten
        self.trainable = trainable
        weights = torch.ones((1, 1, k))
        if trainable:
            self.register_parameter("weights", torch.nn.Parameter(weights))
        else:
            self.register_buffer("weights", weights)

    def forward(self, x: Tensor):  # skipcq: PYL-W0221
        input = x.view(x.size(0), x.size(1), -1)
        kmax = input.topk(k=self.k, dim=2)[0]
        kmax = (kmax * self.weights).mean(dim=2)
        if not self.flatten:
            kmax = kmax.view(kmax.size(0), kmax.size(1), 1, 1)
        return kmax

    def load_state_dict(self, state_dict: Union[Dict[str, Tensor], Dict[str, Tensor]], strict: bool = True):
        if not self.trainable:
            return _IncompatibleKeys([], [])

        super().load_state_dict(state_dict, strict)


class GlobalWeightedAvgPool2d(nn.Module):
    """
    Global Weighted Average Pooling from paper "Global Weighted Average
    Pooling Bridges Pixel-level Localization and Image-level Classification"
    """

    def __init__(self, features: int, flatten=False):
        super().__init__()
        self.conv = nn.Conv2d(features, 1, kernel_size=1, bias=True)
        self.flatten = flatten

    def fscore(self, x):
        m = self.conv(x)
        m = m.sigmoid().exp()
        return m

    def norm(self, x: torch.Tensor):
        return x / x.sum(dim=[2, 3], keepdim=True)

    def forward(self, x):  # skipcq: PYL-W0221
        input_x = x
        x = self.fscore(x)
        x = self.norm(x)
        x = x * input_x
        x = x.sum(dim=[2, 3], keepdim=not self.flatten)
        return x


GWAP = GlobalWeightedAvgPool2d


class RMSPool(nn.Module):
    """
    Root mean square pooling
    """

    def __init__(self):
        super().__init__()
        self.avg_pool = GlobalAvgPool2d()

    def forward(self, x):  # skipcq: PYL-W0221
        x_mean = x.mean(dim=[2, 3])
        avg_pool = self.avg_pool((x - x_mean) ** 2)
        return avg_pool.sqrt()


class MILCustomPoolingModule(nn.Module):
    def __init__(self, in_channels, out_channels, reduction=4):
        super().__init__()
        self.classifier = nn.Conv2d(in_channels, out_channels, kernel_size=1)
        self.weight_generator = nn.Sequential(
            nn.BatchNorm2d(in_channels),
            nn.Conv2d(in_channels, in_channels // reduction, kernel_size=1),
            nn.ReLU(True),
            nn.Conv2d(in_channels // reduction, out_channels, kernel_size=1),
            nn.Sigmoid(),
        )

    def forward(self, x):  # skipcq: PYL-W0221
        weight = self.weight_generator(x)
        loss = self.classifier(x)
        logits = torch.sum(weight * loss, dim=[2, 3]) / (torch.sum(weight, dim=[2, 3]) + 1e-6)
        return logits


class GlobalRankPooling(nn.Module):
    """
    https://arxiv.org/abs/1704.02112
    """

    def __init__(self, num_features, spatial_size, flatten=False):
        super().__init__()
        self.conv = nn.Conv1d(num_features, num_features, spatial_size, groups=num_features)
        self.flatten = flatten

    def forward(self, x: torch.Tensor):  # skipcq: PYL-W0221
        spatial_size = x.size(2) * x.size(3)
        assert spatial_size == self.conv.kernel_size[0], (
            f"Expected spatial size {self.conv.kernel_size[0]}, " f"got {x.size(2)}x{x.size(3)}"
        )

        x = x.view(x.size(0), x.size(1), -1)  # Flatten spatial dimensions
        x_sorted, index = x.topk(spatial_size, dim=2)

        x = self.conv(x_sorted)  # [B, C, 1]

        if self.flatten:
            x = x.squeeze(2)
        return x


class GeneralizedMeanPooling2d(nn.Module):
    """

    https://arxiv.org/pdf/1902.05509v2.pdf
    https://amaarora.github.io/2020/08/30/gempool.html
    """

    def __init__(self, p: float = 3, eps=1e-6, flatten=False):
        super(GeneralizedMeanPooling2d, self).__init__()
        self.p = nn.Parameter(torch.ones(1) * p)
        self.eps = eps
        self.flatten = flatten

    def forward(self, x: Tensor) -> Tensor:
        x = F.adaptive_avg_pool2d(x.clamp_min(self.eps).pow(self.p), output_size=1).pow(1.0 / self.p)
        if self.flatten:
            x = x.view(x.size(0), x.size(1))

        return x

    def __repr__(self):
        return (
            self.__class__.__name__
            + "("
            + "p="
            + "{:.4f}".format(self.p.data.item())
            + ", "
            + "eps="
            + str(self.eps)
            + ")"
        )



from .common import *
from .dpn import *
from .efficient_net import *
from .efficient_net_v2 import *
from .hrnet import *
from .nf_regnet import *
from .nfnet import *
from .nfnet_s import *
from .res2net import *
from .resnet import *



import os
from collections import OrderedDict
from pathlib import Path
from typing import Dict, Union
import math
import torch
from catalyst import utils
from catalyst.callbacks import BaseCheckpointCallback
from catalyst.dl import IRunner, Callback, CallbackOrder


__all__ = [
    "BestMetricCheckpointCallback",
    "HyperParametersCallback",
    "clean_checkpoint",
    "report_checkpoint",
    "sanitize_metric_name",
]

from pytorch_toolbelt.utils.catalyst import get_tensorboard_logger


def clean_checkpoint(src_fname, dst_fname):
    """
    Remove optimizer, scheduler and criterion states from checkpoint
    :param src_fname: Source checkpoint filename
    :param dst_fname: Target checkpoint filename (can be same)
    """
    checkpoint = torch.load(src_fname, map_location="cpu")

    keys = ["criterion_state_dict", "optimizer_state_dict", "scheduler_state_dict"]

    for key in keys:
        if key in checkpoint:
            del checkpoint[key]

    torch.save(checkpoint, dst_fname)


def report_checkpoint(checkpoint: Dict):
    """
    Print checkpoint metrics and epoch number
    :param checkpoint:
    """
    print("Epoch          :", checkpoint["epoch"])

    skip_fields = [
        "_base/lr",
        "_base/momentum",
        "_timers/data_time",
        "_timers/model_time",
        "_timers/batch_time",
        "_timers/_fps",
    ]
    print(
        "Metrics (Train):",
        [
            (k.replace("train_", ""), v)
            for k, v, in checkpoint["epoch_metrics"].items()
            if k not in skip_fields and str.startswith(k, "train_")
        ],
    )
    print(
        "Metrics (Valid):",
        [
            (k.replace("valid_", ""), v)
            for k, v, in checkpoint["epoch_metrics"].items()
            if k not in skip_fields and str.startswith(k, "valid_")
        ],
    )


def sanitize_metric_name(metric_name: str) -> str:
    """
    Replace characters in string that are not path-friendly with underscore
    """
    for s in ["?", "/", "\\", ":", "<", ">", "|", "'", '"', "#"]:
        metric_name = metric_name.replace(s, "_")
    return metric_name


class BestMetricCheckpointCallback(BaseCheckpointCallback):
    """
    Checkpoint callback to save model weights based on user-defined metric value.
    """

    def __init__(
        self,
        target_metric: str,
        target_metric_minimize=False,
        save_n_best: int = 3,
        checkpoints_dir=None,
        metrics_filename: str = "_metrics.json",
    ):
        """
        Args:
            target_metric (str): name of the target metric to monitor.
            target_metric_minimize (bool): define whether metric is minimized.
            save_n_best (int): number of best checkpoint to keep
            checkpoints_dir (str): path to directory where checkpoints will be saved
            metrics_filename (str): filename to save metrics
                in checkpoint folder. Must ends on ``.json`` or ``.yml``
        """
        if checkpoints_dir is None:
            checkpoints_dir = "checkpoints_" + sanitize_metric_name(target_metric)

        super().__init__(metrics_filename=metrics_filename)
        self.main_metric = target_metric
        self.minimize_metric = target_metric_minimize
        self.save_n_best = save_n_best
        self.top_best_metrics = []
        self.epochs_metrics = []
        self.checkpoints_dir = checkpoints_dir
        self.best_main_metric_value = None

    def get_checkpoint_suffix(self, checkpoint: dict) -> str:
        result = f"{checkpoint['stage']}.{checkpoint['epoch']}"
        return result

    def get_metric(self, last_valid_metrics) -> Dict:
        top_best_checkpoints = [
            (Path(filepath).stem, valid_metric) for (filepath, _, valid_metric) in self.top_best_metrics
        ]
        all_epochs_metrics = [
            (f"epoch_{order_index}", valid_metric) for (order_index, valid_metric) in enumerate(self.epochs_metrics)
        ]
        best_valid_metrics = top_best_checkpoints[0][1]
        metrics = OrderedDict(
            [("best", best_valid_metrics)] + [("last", last_valid_metrics)] + top_best_checkpoints + all_epochs_metrics
        )

        self.metrics = metrics
        return self.metrics

    def truncate_checkpoints(self, minimize_metric: bool) -> None:
        def get_proper_sort_key(minimize):
            def get_key(x):
                metric_value = x[1]
                if math.isfinite(metric_value):
                    return metric_value
                else:
                    key = float("+inf") if self.minimize_metric else float("-inf")
                    return key

            return get_key

        self.top_best_metrics = sorted(
            self.top_best_metrics, key=get_proper_sort_key(minimize_metric), reverse=not minimize_metric
        )
        if len(self.top_best_metrics) > self.save_n_best:
            last_item = self.top_best_metrics.pop(-1)
            last_filepath = Path(last_item[0])
            last_filepaths = last_filepath.parent.glob(last_filepath.name.replace(".pth", "*"))
            for filepath in last_filepaths:
                os.remove(filepath)

    def process_checkpoint(
        self, logdir: str, checkpoint: Dict, is_best: bool, main_metric: str = "loss", minimize_metric: bool = True
    ):
        suffix = self.get_checkpoint_suffix(checkpoint)

        exclude = ["criterion", "optimizer", "scheduler"]
        checkpoint = {key: value for key, value in checkpoint.items() if all(z not in key for z in exclude)}
        filepath = utils.save_checkpoint(
            checkpoint=checkpoint,
            logdir=Path(logdir) / Path(self.checkpoints_dir),
            suffix=suffix,
            is_best=is_best,
            is_last=True,
        )

        valid_metrics = checkpoint["valid_metrics"]
        checkpoint_metric = valid_metrics[main_metric]
        metrics_record = (filepath, checkpoint_metric, valid_metrics)
        self.top_best_metrics.append(metrics_record)
        self.epochs_metrics.append(metrics_record)
        self.truncate_checkpoints(minimize_metric=minimize_metric)
        metrics = self.get_metric(valid_metrics)
        self.save_metric(logdir, metrics)

    def on_stage_start(self, state: IRunner):
        self.best_main_metric_value: float = float("+inf") if self.minimize_metric else float("-inf")

    def on_epoch_end(self, state: IRunner):
        if state.stage_name.startswith("infer"):
            return

        valid_metrics = dict(state.valid_metrics)
        epoch_metrics = dict(state.epoch_metrics)

        checkpoint = utils.pack_checkpoint(
            model=state.model,
            criterion=state.criterion,
            optimizer=state.optimizer,
            scheduler=state.scheduler,
            epoch_metrics=epoch_metrics,
            valid_metrics=valid_metrics,
            stage=state.stage_name,
            epoch=state.global_epoch,
            checkpoint_data=state.checkpoint_data,
        )

        main_metric_value = valid_metrics[self.main_metric]
        if self.minimize_metric:
            is_best = math.isfinite(main_metric_value) and main_metric_value < self.best_main_metric_value
        else:
            is_best = math.isfinite(main_metric_value) and main_metric_value > self.best_main_metric_value

        if is_best:
            self.best_main_metric_value = main_metric_value

        self.process_checkpoint(
            logdir=state.logdir,
            checkpoint=checkpoint,
            is_best=is_best,
            main_metric=self.main_metric,
            minimize_metric=self.minimize_metric,
        )

    def on_stage_end(self, state: IRunner):
        print("Top best models:")
        top_best_metrics_str = "\n".join(
            [
                "{filepath}\t{metric:3.4f}".format(filepath=filepath, metric=checkpoint_metric)
                for filepath, checkpoint_metric, _ in self.top_best_metrics
            ]
        )
        print(top_best_metrics_str)

    def save_metric(self, logdir: str, metrics: Dict) -> None:
        utils.save_config(metrics, f"{logdir}/{self.checkpoints_dir}/{self.metrics_filename}")

    def on_exception(self, state: IRunner):
        exception = state.exception
        if not utils.is_exception(exception):
            return

        try:
            valid_metrics = state.valid_metrics
            epoch_metrics = state.epoch_metrics
            checkpoint = utils.pack_checkpoint(
                model=state.model,
                criterion=state.criterion,
                optimizer=state.optimizer,
                scheduler=state.scheduler,
                epoch_metrics=epoch_metrics,
                valid_metrics=valid_metrics,
                stage=state.stage_name,
                epoch=state.epoch,
                checkpoint_data=state.checkpoint_data,
            )
            suffix = self.get_checkpoint_suffix(checkpoint)
            suffix = f"{suffix}.exception_{exception.__class__.__name__}"
            utils.save_checkpoint(
                logdir=Path(f"{state.logdir}/{self.checkpoints_dir}/"),
                checkpoint=checkpoint,
                suffix=suffix,
                is_best=False,
                is_last=False,
            )
            metrics = self.metrics
            metrics[suffix] = valid_metrics
            self.save_metric(state.logdir, metrics)
        except Exception:
            pass


class HyperParametersCallback(Callback):
    """
    Callback that logs hyper-parameters for training session and target metric value.
    Useful for evaluation of several runs in Tensorboard.
    """

    def __init__(self, hparam_dict: Dict[str, Union[str, bool, int, float]]):
        if "stage" in hparam_dict:
            raise KeyError("Key 'stage' is reserved")
        for key, value in hparam_dict.items():
            if not isinstance(value, (str, float, int, bool)):
                raise ValueError(f"Value of key {key} must be either str,float,int,bool. Got {value}")

        super().__init__(CallbackOrder.Metric)
        self.hparam_dict = hparam_dict

    def on_stage_end(self, state: IRunner):
        logger = get_tensorboard_logger(state)

        hparam_dict = self.hparam_dict.copy()
        hparam_dict["stage"] = state.stage_name

        logger.add_hparams(
            hparam_dict=self.hparam_dict,
            metric_dict=state.best_valid_metrics,
        )



from typing import List, Union
from torch import Tensor, nn
import inspect

from .common import SegmentationDecoderModule
from .. import conv1x1, FPNContextBlock, FPNBottleneckBlock

__all__ = ["FPNSumDecoder"]


class FPNSumDecoder(SegmentationDecoderModule):
    """
    Feature pyramid network decoder with summation between intermediate layers:

        Input
        feature_map[0] -> bottleneck[0](feature_map[0]) + upsample(fpn[1]) -> fpn[0]
        feature_map[1] -> bottleneck[1](feature_map[1]) + upsample(fpn[2]) -> fpn[1]
        feature_map[2] -> bottleneck[2](feature_map[2]) + upsample(fpn[3]) -> fpn[2]
        ...
        feature_map[n] -> bottleneck[n](feature_map[n]) + upsample(context) -> fpn[n]
        feature_map[n] -> context_block(feature_map[n]) -> context
    """

    def __init__(
        self,
        feature_maps: List[int],
        channels: int,
        context_block=FPNContextBlock,
        bottleneck_block=FPNBottleneckBlock,
        prediction_block: Union[nn.Identity, conv1x1, nn.Module] = nn.Identity,
        prediction_channels: int = None,
        upsample_block=nn.Upsample,
    ):
        """
        Create a new instance of FPN decoder with summation of consecutive feature maps.
        :param feature_maps: Number of channels in input feature maps (fine to coarse).
            For instance - [64, 256, 512, 2048]
        :param channels: FPN channels
        :param context_block:
        :param bottleneck_block:
        :param prediction_block: Optional prediction block to apply to FPN feature maps before returning from decoder
        :param prediction_channels: Number of prediction channels
        :param upsample_block:
        """
        super().__init__()

        self.context = context_block(feature_maps[-1], channels)

        self.bottlenecks = nn.ModuleList(
            [bottleneck_block(in_channels, channels) for in_channels in reversed(feature_maps)]
        )

        if inspect.isclass(prediction_block) and issubclass(prediction_block, nn.Identity):
            self.outputs = nn.ModuleList([prediction_block() for _ in reversed(feature_maps)])
            self.channels = [channels] * len(feature_maps)
        else:
            self.outputs = nn.ModuleList(
                [prediction_block(channels, prediction_channels) for _ in reversed(feature_maps)]
            )
            self.channels = [prediction_channels] * len(feature_maps)

        if issubclass(upsample_block, nn.Upsample):
            self.upsamples = nn.ModuleList([upsample_block(scale_factor=2) for _ in reversed(feature_maps)])
        else:
            self.upsamples = nn.ModuleList(
                [upsample_block(channels, channels) for in_channels in reversed(feature_maps)]
            )

    def forward(self, feature_maps: List[Tensor]) -> List[Tensor]:
        last_feature_map = feature_maps[-1]
        feature_maps = reversed(feature_maps)

        outputs = []

        fpn = self.context(last_feature_map)

        for feature_map, bottleneck, upsample, output_block in zip(
            feature_maps, self.bottlenecks, self.upsamples, self.outputs
        ):
            fpn = bottleneck(feature_map) + upsample(fpn)
            outputs.append(output_block(fpn))

        # Returns list of tensors in same order as input (fine-to-coarse)
        return outputs[::-1]



from typing import Callable, Optional

import numpy as np
import torch
from catalyst.core import Callback, CallbackOrder
from sklearn.metrics import roc_auc_score
from torch import Tensor

__all__ = ["RocAucMetricCallback"]

from pytorch_toolbelt.utils import to_numpy
from pytorch_toolbelt.utils.distributed import all_gather, is_main_process
from pytorch_toolbelt.utils.catalyst.visualization import get_tensorboard_logger


class RocAucMetricCallback(Callback):
    """
    Roc Auc score metric
    """

    def __init__(
        self,
        outputs_to_probas: Optional[Callable[[Tensor], Tensor]] = torch.sigmoid,
        input_key: str = "targets",
        output_key: str = "logits",
        prefix: str = "roc_auc",
        average="macro",
        ignore_index: Optional[int] = None,
        log_pr_curve: bool = True,
        fix_nans=False,
    ):
        """
        Args:
            input_key: input key to use for accuracy calculation;
                specifies our `y_true`
            output_key: output key to use for accuracy calculation;
                specifies our `y_pred`
            prefix: key for the metric's name
        """
        super().__init__(CallbackOrder.Metric)
        self.prefix = prefix
        self.output_key = output_key
        self.input_key = input_key
        self.ignore_index = ignore_index
        self.outputs_to_probas = outputs_to_probas
        self.y_trues = []
        self.y_preds = []
        self.average = average
        self.log_pr_curve = log_pr_curve
        self.fix_nans = fix_nans

    def on_loader_start(self, state):
        self.y_trues = []
        self.y_preds = []

    @torch.no_grad()
    def on_batch_end(self, runner):
        pred_probas = runner.output[self.output_key].float()
        true_labels = runner.input[self.input_key].float()
        if self.outputs_to_probas:
            pred_probas = self.outputs_to_probas(pred_probas)

        # Aggregate flattened labels
        self.y_trues.extend(to_numpy(true_labels).reshape(-1))
        self.y_preds.extend(to_numpy(pred_probas).reshape(-1))

    def on_loader_end(self, runner):
        y_trues = np.concatenate(all_gather(self.y_trues))
        y_preds = np.concatenate(all_gather(self.y_preds))

        if self.fix_nans:
            y_preds[~np.isfinite(y_preds)] = 0.5

        score = roc_auc_score(y_true=y_trues, y_score=y_preds, average=self.average)
        runner.loader_metrics[self.prefix] = float(score)

        if self.log_pr_curve and is_main_process():
            logger = get_tensorboard_logger(runner)
            logger.add_pr_curve(
                self.prefix, predictions=y_preds, labels=y_trues, global_step=runner.global_epoch, num_thresholds=255
            )



from .common import GenericTimmEncoder, make_n_channel_input_std_conv

__all__ = [
    "NFNetF0SEncoder",
    "NFNetF1SEncoder",
    "NFNetF2SEncoder",
    "NFNetF3SEncoder",
    "NFNetF4SEncoder",
    "NFNetF5SEncoder",
    "NFNetF6SEncoder",
    "NFNetF7SEncoder",
]


class NFNetF0SEncoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nfnet_f0s(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self


class NFNetF1SEncoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        if layers is None:
            layers = [1, 2, 3, 4]
        encoder = nfnet.nfnet_f1s(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self


class NFNetF2SEncoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nfnet_f2s(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self


class NFNetF3SEncoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nfnet_f3s(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self


class NFNetF4SEncoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nfnet_f4s(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self


class NFNetF5SEncoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nfnet_f5s(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self


class NFNetF6SEncoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nfnet_f6s(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self


class NFNetF7SEncoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nfnet_f7s(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self



import pickle
from typing import Any, Dict, List

import torch
from torch import Tensor

import torch.distributed as dist

__all__ = [
    "all_gather",
    "broadcast_from_master",
    "get_rank",
    "get_world_size",
    "is_dist_avail_and_initialized",
    "is_main_process",
    "master_print",
    "reduce_dict_sum",
]


def is_dist_avail_and_initialized() -> bool:
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True


def get_world_size() -> int:
    if not is_dist_avail_and_initialized():
        return 1
    return dist.get_world_size()


def get_rank() -> int:
    if not is_dist_avail_and_initialized():
        return 0
    return dist.get_rank()


def is_main_process() -> bool:
    return get_rank() == 0


def broadcast_from_master(data: Any) -> Any:
    """
    Broadcast data from master node to all other nodes. This may be required when you
    want to compute something only on master node (e.g computational-heavy metric) and
    don't want to vaste CPU of other nodes doing same work simultaneously.

    >>> if is_main_process():
    >>>    result = some_code_to_run(...)
    >>> else:
    >>>    result = None
    >>> # 'result' propagated to all nodes from master
    >>> result = broadcast_from_master(result)

    Args:
        data: Data to be broadcasted from master node (rank 0)

    Returns:
        Data from rank 0 node
    """
    world_size = get_world_size()
    if world_size == 1:
        return data

    local_rank = get_rank()
    storage: Tensor

    if local_rank == 0:
        buffer = pickle.dumps(data)
        storage = torch.ByteStorage.from_buffer(buffer)
        payload = torch.ByteTensor(storage).to("cuda")
        local_size = payload.numel()
    else:
        local_size = 0

    # Propagate target tensor size to all nodes
    local_size = max(all_gather(local_size))
    if local_rank != 0:
        payload = torch.empty((local_size,), dtype=torch.uint8, device="cuda")

    dist.broadcast(payload, 0)
    buffer = payload.cpu().numpy().tobytes()
    return pickle.loads(buffer)


def all_gather(data: Any) -> List[Any]:
    """
    Run all_gather on arbitrary picklable data (not necessarily tensors)
    Args:
        data: any picklable object
    Returns:
        list[data]: list of data gathered from each rank
    """
    world_size = get_world_size()
    if world_size == 1:
        return [data]

    # serialized to a Tensor
    buffer = pickle.dumps(data)
    storage = torch.ByteStorage.from_buffer(buffer)
    tensor = torch.ByteTensor(storage).to("cuda")

    # obtain Tensor size of each rank
    local_size = torch.tensor([tensor.numel()], device="cuda")
    size_list = [torch.tensor([0], device="cuda") for _ in range(world_size)]
    dist.all_gather(size_list, local_size)
    size_list = [int(size.item()) for size in size_list]
    max_size = max(size_list)

    # receiving Tensor from all ranks
    # we pad the tensor because torch all_gather does not support
    # gathering tensors of different shapes
    tensor_list = []
    for _ in size_list:
        tensor_list.append(torch.empty((max_size,), dtype=torch.uint8, device="cuda"))
    if local_size != max_size:
        padding = torch.empty(size=(max_size - local_size,), dtype=torch.uint8, device="cuda")
        tensor = torch.cat((tensor, padding), dim=0)
    dist.all_gather(tensor_list, tensor)

    data_list = []
    for size, tensor in zip(size_list, tensor_list):
        buffer = tensor.cpu().numpy().tobytes()[:size]
        data_list.append(pickle.loads(buffer))

    return data_list


def reduce_dict_sum(input_dict: Dict[Any, Any]) -> Dict[Any, Any]:
    """
    Reduce the values in the dictionary from all processes so that all processes
    have the sum of the all values. Returns a dict with the same fields as
    input_dict, after reduction.

    Note: This function can work on defaultdict(list) and will effectively
    concatenate multiple dictionaries into single one, thanks to that summation
    of lists works as a concatenation.
    Args:
        input_dict (dict): all the values will be reduced
    """
    world_size = get_world_size()
    if world_size < 2:
        return input_dict

    with torch.no_grad():
        reduced_dict = {}
        for x in all_gather(input_dict):
            for key, value in x.items():
                if key in reduced_dict:
                    reduced_dict[key] += value
                else:
                    reduced_dict[key] = value
    return reduced_dict


def master_print(*args, **kwargs) -> None:
    """
    Drop-in replacement for built-in `print` function that prints only on master node

    Args:
        *args:
        **kwargs:

    Returns:
        None
    """
    if is_main_process():
        print(*args, **kwargs)



"""Wrappers for different backbones for models that follows Encoder-Decoder architecture.

Encodes listed here provides easy way to swap backbone of classification/segmentation/detection model.
"""

from collections import OrderedDict

from torch import nn
from torchvision.models import resnet50, resnet34, resnet18, resnet101, resnet152

from .common import EncoderModule, _take, make_n_channel_input

__all__ = [
    "ResnetEncoder",
    "Resnet18Encoder",
    "Resnet34Encoder",
    "Resnet50Encoder",
    "Resnet101Encoder",
    "Resnet152Encoder",
]


class ResnetEncoder(EncoderModule):
    def __init__(self, resnet, filters, strides, layers=None):
        if layers is None:
            layers = [1, 2, 3, 4]
        super().__init__(filters, strides, layers)

        self.layer0 = nn.Sequential(OrderedDict([("conv0", resnet.conv1), ("bn0", resnet.bn1), ("act0", resnet.relu)]))
        self.maxpool = resnet.maxpool

        self.layer1 = resnet.layer1
        self.layer2 = resnet.layer2
        self.layer3 = resnet.layer3
        self.layer4 = resnet.layer4

    @property
    def encoder_layers(self):
        return [self.layer0, self.layer1, self.layer2, self.layer3, self.layer4]

    def forward(self, x):
        output_features = []
        for layer in self.encoder_layers:
            output = layer(x)
            output_features.append(output)

            if layer == self.layer0:
                # Fist maxpool operator is not a part of layer0 because we want that layer0 output to have stride of 2
                output = self.maxpool(output)
            x = output

        # Return only features that were requested
        return _take(output_features, self._layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.layer0.conv0 = make_n_channel_input(self.layer0.conv0, input_channels, mode=mode, **kwargs)
        return self


class Resnet18Encoder(ResnetEncoder):
    def __init__(self, pretrained=True, layers=None):
        super().__init__(resnet18(pretrained=pretrained), [64, 64, 128, 256, 512], [2, 4, 8, 16, 32], layers)


class Resnet34Encoder(ResnetEncoder):
    def __init__(self, pretrained=True, layers=None):
        super().__init__(resnet34(pretrained=pretrained), [64, 64, 128, 256, 512], [2, 4, 8, 16, 32], layers)


class Resnet50Encoder(ResnetEncoder):
    def __init__(self, pretrained=True, layers=None):
        super().__init__(resnet50(pretrained=pretrained), [64, 256, 512, 1024, 2048], [2, 4, 8, 16, 32], layers)


class Resnet101Encoder(ResnetEncoder):
    def __init__(self, pretrained=True, layers=None):
        super().__init__(resnet101(pretrained=pretrained), [64, 256, 512, 1024, 2048], [2, 4, 8, 16, 32], layers)


class Resnet152Encoder(ResnetEncoder):
    def __init__(self, pretrained=True, layers=None):
        super().__init__(resnet152(pretrained=pretrained), [64, 256, 512, 1024, 2048], [2, 4, 8, 16, 32], layers)



from typing import Any, Dict
from .support import pytorch_toolbelt_deprecated

__all__ = ["maybe_eval", "without", "load_yaml"]


def maybe_eval(x: str) -> Any:
    if isinstance(x, str):
        if x.startswith("$"):
            return eval(x[1:])
        return x
    elif isinstance(x, list):
        return list(map(maybe_eval, x))
    else:
        return x


def without(dictionary: Dict, key: str) -> Dict:
    """
    Return copy of dictionary without given key.

    Args:
        dictionary: Input dictionary
        key: Key to remove

    Returns:
        Always returns new dictionary even without given key
    """
    if isinstance(key, str):
        key = {key}
    return dict((k, v) for (k, v) in dictionary.items() if k not in key)


@pytorch_toolbelt_deprecated("This method is deprecated. Please use OmegaConf")
def load_yaml(stream: Any):
    """
    Parse the first YAML document in a stream and produce the corresponding Python object.
    This function support parsing float values like `1e-4`.

    Implementation credit: https://stackoverflow.com/questions/30458977/yaml-loads-5e-6-as-string-and-not-a-number

    Args:
        stream:
    """
    import yaml
    import re

    loader = yaml.SafeLoader
    loader.add_implicit_resolver(
        "tag:yaml.org,2002:float",
        re.compile(
            """^(?:
         [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
        |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
        |\\.[0-9_]+(?:[eE][-+][0-9]+)?
        |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
        |[-+]?\\.(?:inf|Inf|INF)
        |\\.(?:nan|NaN|NAN))$""",
            re.X,
        ),
        list("-+0123456789."),
    )

    return yaml.load(stream, Loader=loader)



from typing import Optional, List

import torch
from torch import nn, Tensor
from math import hypot

__all__ = [
    "bilinear_upsample_initializer",
    "icnr_init",
    "DepthToSpaceUpsample2d",
    "BilinearAdditiveUpsample2d",
    "DeconvolutionUpsample2d",
    "ResidualDeconvolutionUpsample2d",
]


def bilinear_upsample_initializer(x):
    cc = x.size(2) // 2
    cr = x.size(3) // 2

    for i in range(x.size(2)):
        for j in range(x.size(3)):
            x[..., i, j] = hypot(cc - i, cr - j)

    y = 1 - x / x.sum(dim=(2, 3), keepdim=True)
    y = y / y.sum(dim=(2, 3), keepdim=True)
    return y


def icnr_init(tensor: torch.Tensor, upscale_factor=2, initializer=nn.init.kaiming_normal):
    """Fill the input Tensor or Variable with values according to the method
    described in "Checkerboard artifact free sub-pixel convolution"
    - Andrew Aitken et al. (2017), this inizialization should be used in the
    last convolutional layer before a PixelShuffle operation
    Args:
        tensor: an n-dimensional torch.Tensor or autograd.Variable
        upscale_factor: factor to increase spatial resolution by
        initializer: inizializer to be used for sub_kernel inizialization
    Examples:
        >>> upscale = 8
        >>> num_classes = 10
        >>> previous_layer_features = Variable(torch.Tensor(8, 64, 32, 32))
        >>> conv_shuffle = Conv2d(64, num_classes * (upscale ** 2), 3, padding=1, bias=0)
        >>> ps = PixelShuffle(upscale)
        >>> kernel = ICNR(conv_shuffle.weight, scale_factor=upscale)
        >>> conv_shuffle.weight.data.copy_(kernel)
        >>> output = ps(conv_shuffle(previous_layer_features))
        >>> print(output.shape)
        torch.Size([8, 10, 256, 256])
    .. _Checkerboard artifact free sub-pixel convolution:
        https://arxiv.org/abs/1707.02937
    """
    new_shape = [int(tensor.shape[0] / (upscale_factor**2))] + list(tensor.shape[1:])
    subkernel = torch.zeros(new_shape)
    subkernel = initializer(subkernel)
    subkernel = subkernel.transpose(0, 1)

    subkernel = subkernel.contiguous().view(subkernel.shape[0], subkernel.shape[1], -1)

    kernel = subkernel.repeat(1, 1, upscale_factor**2)

    transposed_shape = [tensor.shape[1]] + [tensor.shape[0]] + list(tensor.shape[2:])
    kernel = kernel.contiguous().view(transposed_shape)

    kernel = kernel.transpose(0, 1)
    return kernel


class DepthToSpaceUpsample2d(nn.Module):
    """
    NOTE: This block is not fully ready yet. Need to figure out how to correctly initialize
    default weights to have bilinear upsample identical to OpenCV results

    https://github.com/pytorch/pytorch/pull/5429
    https://arxiv.org/ftp/arxiv/papers/1707/1707.02937.pdf
    """

    def __init__(self, in_channels: int, out_channels: int, scale_factor: int = 2):
        super().__init__()
        n = 2**scale_factor
        self.conv = nn.Conv2d(in_channels, out_channels * n, kernel_size=3, padding=1, bias=False)
        self.out_channels = out_channels
        self.shuffle = nn.PixelShuffle(upscale_factor=scale_factor)

        # with torch.no_grad():
        #     self.conv.weight.data = icnr_init(
        #         self.conv.weight, upscale_factor=scale_factor, initializer=bilinear_upsample_initializer
        #     )

    def forward(self, x: Tensor) -> Tensor:  # skipcq: PYL-W0221
        x = self.shuffle(self.conv(x))
        return x


class BilinearAdditiveUpsample2d(nn.Module):
    """
    https://arxiv.org/abs/1707.05847
    """

    def __init__(self, in_channels: int, scale_factor: int = 2, n: int = 4):
        super().__init__()
        if in_channels % n != 0:
            raise ValueError(f"Number of input channels ({in_channels})must be divisable by n ({n})")

        self.in_channels = in_channels
        self.out_channels = in_channels // n
        self.upsample = nn.UpsamplingBilinear2d(scale_factor=scale_factor)
        self.n = n

    def forward(self, x: Tensor) -> Tensor:  # skipcq: PYL-W0221
        x = self.upsample(x)
        n, c, h, w = x.size()
        x = x.reshape(n, c // self.n, self.n, h, w).mean(2)
        return x


class DeconvolutionUpsample2d(nn.Module):
    def __init__(self, in_channels: int, n=4):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = in_channels // n
        self.conv = nn.ConvTranspose2d(in_channels, in_channels // n, kernel_size=3, padding=1, stride=2)

    def forward(self, x: Tensor, output_size: Optional[List[int]] = None) -> Tensor:  # skipcq: PYL-W0221
        return self.conv(x, output_size=output_size)


class ResidualDeconvolutionUpsample2d(nn.Module):
    def __init__(self, in_channels, scale_factor=2, n=4):
        if scale_factor != 2:
            raise NotImplementedError("Scale factor other than 2 is not implemented")
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = in_channels // n
        self.conv = nn.ConvTranspose2d(
            in_channels, in_channels // n, kernel_size=3, padding=1, stride=scale_factor, output_padding=1
        )
        self.residual = BilinearAdditiveUpsample2d(in_channels, scale_factor=scale_factor, n=n)
        self.init_weights()

    def forward(self, x: Tensor) -> Tensor:  # skipcq: PYL-W0221
        residual_up = self.residual(x)
        return self.conv(x, output_size=residual_up.size()) + residual_up

    def init_weights(self):
        torch.nn.init.kaiming_normal_(self.conv.weight, nonlinearity="linear")
        torch.nn.init.zeros_(self.conv.bias)



from torch import nn, Tensor

__all__ = ["DecoderModule", "SegmentationDecoderModule"]

from typing import List


class DecoderModule(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, feature_maps: List[Tensor]) -> List[Tensor]:  # skipcq: PYL-W0221
        raise NotImplementedError

    def set_trainable(self, trainable):
        for param in self.parameters():
            param.requires_grad = bool(trainable)


class SegmentationDecoderModule(DecoderModule):
    """
    A placeholder for future. Indicates sub-class decoders are suitable for segmentation tasks
    """

    pass



from typing import List

import torch
import torch.nn.functional as F
from torch import nn


def conv3x3_bn_relu(in_planes, out_planes, stride=1):
    """
    3x3 convolution + BN + relu
    """
    return nn.Sequential(
        nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False),
        nn.BatchNorm2d(out_planes),
        nn.ReLU(inplace=True),
    )


class UPerNet(nn.Module):
    def __init__(self, output_filters: List[int], num_classes=150, pool_scales=(1, 2, 3, 6), fpn_dim=256):
        super(UPerNet, self).__init__()

        last_fm_dim = output_filters[-1]

        # PPM Module
        self.ppm_pooling = []
        self.ppm_conv = []

        for scale in pool_scales:
            self.ppm_pooling.append(nn.AdaptiveAvgPool2d(scale))
            self.ppm_conv.append(
                nn.Sequential(
                    nn.Conv2d(last_fm_dim, 512, kernel_size=1, bias=False), nn.BatchNorm2d(512), nn.ReLU(inplace=True)
                )
            )
        self.ppm_pooling = nn.ModuleList(self.ppm_pooling)
        self.ppm_conv = nn.ModuleList(self.ppm_conv)
        self.ppm_last_conv = conv3x3_bn_relu(last_fm_dim + len(pool_scales) * 512, fpn_dim, 1)

        # FPN Module
        self.fpn_in = []
        for fpn_inplane in output_filters[:-1]:  # skip the top layer
            self.fpn_in.append(
                nn.Sequential(
                    nn.Conv2d(fpn_inplane, fpn_dim, kernel_size=1, bias=False),
                    nn.BatchNorm2d(fpn_dim),
                    nn.ReLU(inplace=True),
                )
            )
        self.fpn_in = nn.ModuleList(self.fpn_in)

        self.fpn_out = []
        for i in range(len(output_filters) - 1):  # skip the top layer
            self.fpn_out.append(nn.Sequential(conv3x3_bn_relu(fpn_dim, fpn_dim, 1)))
        self.fpn_out = nn.ModuleList(self.fpn_out)

        self.conv_last = nn.Sequential(
            conv3x3_bn_relu(len(output_filters) * fpn_dim, fpn_dim, 1), nn.Conv2d(fpn_dim, num_classes, kernel_size=1)
        )

    def forward(self, feature_maps):
        last_fm = feature_maps[-1]

        input_size = last_fm.size()
        ppm_out = [last_fm]
        for pool_scale, pool_conv in zip(self.ppm_pooling, self.ppm_conv):
            ppm_out.append(
                pool_conv(
                    F.interpolate(
                        pool_scale(last_fm), (input_size[2], input_size[3]), mode="bilinear", align_corners=False
                    )
                )
            )
        ppm_out = torch.cat(ppm_out, 1)
        f = self.ppm_last_conv(ppm_out)

        fpn_feature_list = [f]
        for i in reversed(range(len(feature_maps) - 1)):
            conv_x = feature_maps[i]
            conv_x = self.fpn_in[i](conv_x)  # lateral branch

            f = F.interpolate(f, size=conv_x.size()[2:], mode="bilinear", align_corners=False)  # top-down branch
            f = conv_x + f

            fpn_feature_list.append(self.fpn_out[i](f))

        fpn_feature_list.reverse()  # [P2 - P5]
        output_size = fpn_feature_list[0].size()[2:]
        fusion_list = [fpn_feature_list[0]]
        for i in range(1, len(fpn_feature_list)):
            fusion_list.append(F.interpolate(fpn_feature_list[i], output_size, mode="bilinear", align_corners=False))

        fusion_out = torch.cat(fusion_list, 1)
        x = self.conv_last(fusion_out)
        return x



import math
import warnings
import torch

from typing import List, Union
from torch import Tensor, nn

from ..common import EncoderModule, _take_ints, _take_tensors

__all__ = ["GenericTimmEncoder", "make_n_channel_input_std_conv"]


class GenericTimmEncoder(EncoderModule):
    def __init__(self, timm_encoder: Union[nn.Module, str], layers: List[int] = None, pretrained=True):
        strides = []
        channels = []
        default_layers = []
        if isinstance(timm_encoder, str):
            import timm.models.factory

            timm_encoder = timm.models.factory.create_model(timm_encoder, features_only=True, pretrained=pretrained)

        for i, fi in enumerate(timm_encoder.feature_info):
            strides.append(fi["reduction"])
            channels.append(fi["num_chs"])
            default_layers.append(i)

        if layers is None:
            layers = default_layers

        super().__init__(channels, strides, layers)
        self.encoder = timm_encoder

    def forward(self, x: Tensor) -> List[Tensor]:
        all_feature_maps = self.encoder(x)
        return _take_tensors(all_feature_maps, self._layers)


def make_n_channel_input_std_conv(conv: nn.Module, in_channels: int, mode="auto", **kwargs) -> nn.Module:
    """
    Return the same convolution class but with desired number of channels

    Args:
        conv: Input nn.Conv2D object to copy settings/weights from
        in_channels: Desired number of input channels
        mode:
        **kwargs: Optional overrides for Conv2D parameters
    """
    conv_cls = conv.__class__

    if conv.in_channels == in_channels:
        warnings.warn("make_n_channel_input call is spurious")
        return conv

    new_conv = conv_cls(
        in_channels,
        out_channels=conv.out_channels,
        kernel_size=kwargs.get("kernel_size", conv.kernel_size),
        stride=kwargs.get("stride", conv.stride),
        padding=kwargs.get("padding", conv.padding),
        dilation=kwargs.get("dilation", conv.dilation),
        groups=kwargs.get("groups", conv.groups),
        bias=kwargs.get("bias", conv.bias is not None),
        eps=kwargs.get("eps", conv.eps),
    )

    w = conv.weight
    if in_channels > conv.in_channels:
        n = math.ceil(in_channels / float(conv.in_channels))
        w = torch.cat([w] * n, dim=1)
        w = w[:, :in_channels, ...]
        new_conv.weight = nn.Parameter(w, requires_grad=True)
    else:
        w = w[:, 0:in_channels, ...]
        new_conv.weight = nn.Parameter(w, requires_grad=True)

    return new_conv



from collections import OrderedDict

import torch
from torch import nn
from torch.nn import functional as F

from .identity import Identity

__all__ = [
    "ABN",
    "ACT_ELU",
    "ACT_GELU",
    "ACT_HARD_SIGMOID",
    "ACT_HARD_SWISH",
    "ACT_LEAKY_RELU",
    "ACT_MISH",
    "ACT_MISH_NAIVE",
    "ACT_NONE",
    "ACT_RELU",
    "ACT_RELU6",
    "ACT_SELU",
    "ACT_SIGMOID",
    "ACT_SILU",
    "ACT_SOFTPLUS",
    "ACT_SWISH",
    "ACT_SWISH_NAIVE",
    "AGN",
    "HardSigmoid",
    "HardSwish",
    "Mish",
    "MishNaive",
    "Swish",
    "SwishNaive",
    "get_activation_block",
    "hard_sigmoid",
    "hard_swish",
    "instantiate_activation_block",
    "mish",
    "mish_naive",
    "sanitize_activation_name",
    "swish",
    "swish_naive",
]

# Activation names
ACT_CELU = "celu"
ACT_ELU = "elu"
ACT_GELU = "gelu"
ACT_GLU = "glu"
ACT_HARD_SIGMOID = "hard_sigmoid"
ACT_HARD_SWISH = "hard_swish"
ACT_LEAKY_RELU = "leaky_relu"
ACT_MISH = "mish"
ACT_MISH_NAIVE = "mish_naive"
ACT_NONE = "none"
ACT_PRELU = "prelu"
ACT_RELU = "relu"
ACT_RELU6 = "relu6"
ACT_SELU = "selu"
ACT_SIGMOID = "sigmoid"
ACT_SILU = "silu"
ACT_SOFTPLUS = "softplus"
ACT_SWISH = "swish"
ACT_SWISH_NAIVE = "swish_naive"

# This version reduces memory overhead of Swish during training by
# recomputing torch.sigmoid(x) in backward instead of saving it.


@torch.jit.script
def swish_jit_fwd(x):
    return x.mul(torch.sigmoid(x))


@torch.jit.script
def swish_jit_bwd(x, grad_output):
    x_sigmoid = torch.sigmoid(x)
    return grad_output * (x_sigmoid * (1 + x * (1 - x_sigmoid)))


class SwishFunction(torch.autograd.Function):
    """
    Memory efficient Swish implementation.

    Credit:
        https://blog.ceshine.net/post/pytorch-memory-swish/
        https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/layers/activations_jit.py

    """

    @staticmethod
    def forward(ctx, x):
        ctx.save_for_backward(x)
        return swish_jit_fwd(x)

    @staticmethod
    def backward(ctx, grad_output):
        x = ctx.saved_tensors[0]
        return swish_jit_bwd(x, grad_output)


def swish(x):
    return SwishFunction.apply(x)


def swish_naive(x):
    return x * x.sigmoid()


def mish_naive(input):
    """
    Apply the mish function element-wise:
    mish(x) = x * tanh(softplus(x)) = x * tanh(ln(1 + exp(x)))
    See additional documentation for mish class.
    Credit: https://github.com/digantamisra98/Mish
    """
    return input * torch.tanh(F.softplus(input))


class MishNaive(nn.Module):
    def __init__(self, inplace=False):
        super().__init__()

    def forward(self, x):
        return mish_naive(x)


@torch.jit.script
def mish_jit_fwd(x):
    return x.mul(torch.tanh(F.softplus(x)))


@torch.jit.script
def mish_jit_bwd(x, grad_output):
    x_sigmoid = torch.sigmoid(x)
    x_tanh_sp = F.softplus(x).tanh()
    return grad_output.mul(x_tanh_sp + x * x_sigmoid * (1 - x_tanh_sp * x_tanh_sp))


class MishFunction(torch.autograd.Function):
    """
    Mish: A Self Regularized Non-Monotonic Neural Activation Function - https://arxiv.org/abs/1908.08681
    A memory efficient, jit scripted variant of Mish
    """

    @staticmethod
    def forward(ctx, x):
        ctx.save_for_backward(x)
        return mish_jit_fwd(x)

    @staticmethod
    def backward(ctx, grad_output):
        x = ctx.saved_tensors[0]
        return mish_jit_bwd(x, grad_output)


def mish(x):
    """
    Apply the mish function element-wise:
    mish(x) = x * tanh(softplus(x)) = x * tanh(ln(1 + exp(x)))
    See additional documentation for mish class.
    Credit: https://github.com/digantamisra98/Mish
    """
    return MishFunction.apply(x)


class Mish(nn.Module):
    """
    Applies the mish function element-wise:
    mish(x) = x * tanh(softplus(x)) = x * tanh(ln(1 + exp(x)))
    Shape:
        - Input: (N, *) where * means, any number of additional
          dimensions
        - Output: (N, *), same shape as the input
    Examples:
        >>> m = Mish()
        >>> input = torch.randn(2)
        >>> output = m(input)

    Credit: https://github.com/digantamisra98/Mish
    """

    def __init__(self, inplace=False):
        """
        Init method.
        :param inplace: Not used, exists only for compatibility
        """
        super().__init__()

    def forward(self, input):
        """
        Forward pass of the function.
        """
        return mish(input)


def hard_sigmoid(x, inplace=False):
    return F.relu6(x + 3, inplace) / 6


def hard_swish(x, inplace=False):
    return x * hard_sigmoid(x, inplace)


class HardSigmoid(nn.Module):
    def __init__(self, inplace=False):
        super(HardSigmoid, self).__init__()
        self.inplace = inplace

    def forward(self, x):
        return hard_sigmoid(x, inplace=self.inplace)


class SwishNaive(nn.Module):
    def forward(self, input_tensor):
        return swish_naive(input_tensor)


class Swish(nn.Module):
    def __init__(self, inplace=False):
        super(Swish, self).__init__()

    def forward(self, input_tensor):
        return swish(input_tensor)


class HardSwish(nn.Module):
    def __init__(self, inplace=False):
        super(HardSwish, self).__init__()
        self.inplace = inplace

    def forward(self, x):
        return hard_swish(x, inplace=self.inplace)


def get_activation_block(activation_name: str):
    ACTIVATIONS = {
        ACT_CELU: nn.CELU,
        ACT_ELU: nn.ELU,
        ACT_GELU: nn.GELU,
        ACT_GLU: nn.GLU,
        ACT_HARD_SIGMOID: HardSigmoid,
        ACT_HARD_SWISH: HardSwish,
        ACT_LEAKY_RELU: nn.LeakyReLU,
        ACT_MISH: Mish,
        ACT_MISH_NAIVE: MishNaive,
        ACT_NONE: Identity,
        ACT_PRELU: nn.PReLU,
        ACT_RELU6: nn.ReLU6,
        ACT_RELU: nn.ReLU,
        ACT_SELU: nn.SELU,
        ACT_SILU: nn.SiLU,
        ACT_SOFTPLUS: nn.Softplus,
        ACT_SWISH: Swish,
        ACT_SWISH_NAIVE: SwishNaive,
        ACT_SIGMOID: nn.Sigmoid,
    }

    return ACTIVATIONS[activation_name.lower()]


def instantiate_activation_block(activation_name: str, **kwargs) -> nn.Module:
    block = get_activation_block(activation_name)

    act_params = {}

    if "inplace" in kwargs and activation_name in {
        ACT_RELU,
        ACT_RELU6,
        ACT_LEAKY_RELU,
        ACT_SELU,
        ACT_SILU,
        ACT_CELU,
        ACT_ELU,
    }:
        act_params["inplace"] = kwargs["inplace"]

    if "slope" in kwargs and activation_name in {ACT_LEAKY_RELU}:
        act_params["negative_slope"] = kwargs["slope"]

    if activation_name == ACT_PRELU:
        act_params["num_parameters"] = kwargs["num_parameters"]

    return block(**act_params)


def sanitize_activation_name(activation_name: str) -> str:
    """
    Return reasonable activation name for initialization in `kaiming_uniform_` for hipster activations
    """
    if activation_name in {ACT_MISH, ACT_SWISH, ACT_SWISH_NAIVE, ACT_MISH_NAIVE}:
        return ACT_LEAKY_RELU

    return activation_name


def ABN(
    num_features: int,
    eps=1e-5,
    momentum=0.1,
    affine=True,
    track_running_stats=True,
    activation=ACT_RELU,
    slope=0.01,
    inplace=True,
):
    bn = nn.BatchNorm2d(
        num_features, eps=eps, momentum=momentum, affine=affine, track_running_stats=track_running_stats
    )
    act = instantiate_activation_block(activation, inplace=inplace, slope=slope)
    return nn.Sequential(OrderedDict([("bn", bn), (activation, act)]))


def AGN(num_features: int, num_groups: int, eps=1e-5, affine=True, activation=ACT_RELU, slope=0.01, inplace=True):
    gn = nn.GroupNorm(num_channels=num_features, num_groups=num_groups, eps=eps, affine=affine)
    act = instantiate_activation_block(activation, inplace=inplace, slope=slope)
    return nn.Sequential(OrderedDict([("gn", gn), (activation, act)]))



from typing import List, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from .common import DecoderModule
from ..activations import get_activation_block, ACT_RELU

__all__ = ["BiFPNDecoder", "BiFPNBlock", "BiFPNConvBlock", "BiFPNDepthwiseConvBlock"]


class BiFPNDepthwiseConvBlock(nn.Module):
    """
    Depthwise seperable convolution.


    """

    def __init__(self, in_channels, out_channels, kernel_size=1, stride=1, padding=0, dilation=1, act=nn.ReLU):
        super(BiFPNDepthwiseConvBlock, self).__init__()
        self.depthwise = nn.Conv2d(
            in_channels, in_channels, kernel_size, stride, padding, dilation, groups=in_channels, bias=False
        )
        self.pointwise = nn.Conv2d(
            in_channels, out_channels, kernel_size=1, stride=1, padding=0, dilation=1, groups=1, bias=False
        )

        self.bn = nn.BatchNorm2d(out_channels, momentum=0.9997, eps=4e-5)
        self.act = act(inplace=True)

    def forward(self, x: Tensor) -> Tensor:
        x = self.depthwise(x)
        x = self.pointwise(x)
        x = self.bn(x)
        return self.act(x)


class BiFPNConvBlock(nn.Module):
    """
    Convolution block with Batch Normalization and ReLU activation.

    """

    def __init__(self, in_channels, out_channels, kernel_size=1, stride=1, padding=0, act=nn.ReLU, dilation=1):
        super(BiFPNConvBlock, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, stride=stride, padding=padding, bias=False)
        self.bn = nn.BatchNorm2d(out_channels, momentum=0.9997, eps=4e-5)
        self.act = act(inplace=True)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        x = self.bn(x)
        return self.act(x)


class BiFPNBlock(nn.Module):
    """
    Bi-directional Feature Pyramid Network
    """

    def __init__(self, feature_size: int = 64, epsilon: float = 0.0001, act=nn.ReLU):
        super(BiFPNBlock, self).__init__()
        self.epsilon = epsilon

        self.p3_td = BiFPNDepthwiseConvBlock(feature_size, feature_size, act=act)
        self.p4_td = BiFPNDepthwiseConvBlock(feature_size, feature_size, act=act)
        self.p5_td = BiFPNDepthwiseConvBlock(feature_size, feature_size, act=act)
        self.p6_td = BiFPNDepthwiseConvBlock(feature_size, feature_size, act=act)

        self.p4_out = BiFPNDepthwiseConvBlock(feature_size, feature_size, act=act)
        self.p5_out = BiFPNDepthwiseConvBlock(feature_size, feature_size, act=act)
        self.p6_out = BiFPNDepthwiseConvBlock(feature_size, feature_size, act=act)
        self.p7_out = BiFPNDepthwiseConvBlock(feature_size, feature_size, act=act)

        self.w1 = nn.Parameter(torch.Tensor(2, 4), requires_grad=True)
        self.w1_relu = nn.ReLU()
        self.w2 = nn.Parameter(torch.Tensor(3, 4), requires_grad=True)
        self.w2_relu = nn.ReLU()

        torch.nn.init.constant_(self.w1, 1)
        torch.nn.init.constant_(self.w2, 1)

    def forward(self, inputs: List[Tensor]) -> List[Tensor]:
        p3_x, p4_x, p5_x, p6_x, p7_x = inputs

        # Calculate Top-Down Pathway
        w1 = self.w1_relu(self.w1)
        w1 = w1 / (torch.sum(w1, dim=0) + self.epsilon)

        w2 = self.w2_relu(self.w2)
        w2 = w2 / (torch.sum(w2, dim=0) + self.epsilon)

        p7_td = p7_x
        p6_td = self.p6_td(w1[0, 0] * p6_x + w1[1, 0] * F.interpolate(p7_td, size=p6_x.size()[2:]))
        p5_td = self.p5_td(w1[0, 1] * p5_x + w1[1, 1] * F.interpolate(p6_td, size=p5_x.size()[2:]))
        p4_td = self.p4_td(w1[0, 2] * p4_x + w1[1, 2] * F.interpolate(p5_td, size=p4_x.size()[2:]))
        p3_td = self.p3_td(w1[0, 3] * p3_x + w1[1, 3] * F.interpolate(p4_td, size=p3_x.size()[2:]))

        # Calculate Bottom-Up Pathway
        p3_out = p3_td
        p4_out = self.p4_out(
            w2[0, 0] * p4_x + w2[1, 0] * p4_td + w2[2, 0] * F.interpolate(p3_out, size=p4_x.size()[2:])
        )
        p5_out = self.p5_out(
            w2[0, 1] * p5_x + w2[1, 1] * p5_td + w2[2, 1] * F.interpolate(p4_out, size=p5_x.size()[2:])
        )
        p6_out = self.p6_out(
            w2[0, 2] * p6_x + w2[1, 2] * p6_td + w2[2, 2] * F.interpolate(p5_out, size=p6_x.size()[2:])
        )
        p7_out = self.p7_out(
            w2[0, 3] * p7_x + w2[1, 3] * p7_td + w2[2, 3] * F.interpolate(p6_out, size=p7_x.size()[2:])
        )

        return [p3_out, p4_out, p5_out, p6_out, p7_out]


class BiFPNDecoder(DecoderModule):
    """
    BiFPN decoder

    Expects input of three feature maps

    Reference: https://arxiv.org/abs/1911.09070
    """

    def __init__(
        self, feature_maps: List[int], strides: List[int], channels: int = 64, num_layers: int = 2, activation=ACT_RELU
    ):
        super(BiFPNDecoder, self).__init__()
        act = get_activation_block(activation)
        if len(feature_maps) != 3:
            raise ValueError("Number of input feature maps must be equal 3")
        self.p3 = nn.Conv2d(feature_maps[0], channels, kernel_size=(1, 1))
        self.p4 = nn.Conv2d(feature_maps[1], channels, kernel_size=(1, 1))
        self.p5 = nn.Conv2d(feature_maps[2], channels, kernel_size=(1, 1))

        # p6 is obtained via a 3x3 stride-2 conv on C5
        self.p6 = nn.Conv2d(feature_maps[2], channels, kernel_size=(3, 3), stride=(2, 2), padding=1)

        # p7 is computed by applying ReLU followed by a 3x3 stride-2 conv on p6
        self.p7 = BiFPNConvBlock(channels, channels, kernel_size=3, stride=2, padding=1, act=act)

        bifpns: List[BiFPNBlock] = []
        for _ in range(num_layers):
            bifpns.append(BiFPNBlock(channels, act=act))
        self.bifpn = nn.ModuleList(bifpns)

        self._channels = [channels] * 5
        self._strides = tuple(strides) + (strides[-1] * 2, strides[-1] * 4)

    @property
    def channels(self):
        return self._channels

    @property
    def strides(self):
        return self._strides

    def forward(self, inputs: List[Tensor]) -> List[Tensor]:
        c3, c4, c5 = inputs

        # Calculate the input column of BiFPN
        p3_x = self.p3(c3)
        p4_x = self.p4(c4)
        p5_x = self.p5(c5)
        p6_x = self.p6(c5)
        p7_x = self.p7(p6_x)

        features = p3_x, p4_x, p5_x, p6_x, p7_x
        for bifpn in self.bifpn:
            features = bifpn(features)
        return features



from typing import Optional, Iterator, Dict
from torch import nn

__all__ = ["get_lr_decay_parameters", "get_optimizable_parameters", "freeze_model"]


def get_lr_decay_parameters(model: nn.Module, learning_rate: float, lr_multipliers: Dict[str, float]):
    """
    Create different parameter groups with different settings.

    Args:
        parameters:
        learning_rate:
        groups: {"encoder": 0.1 ,"encoder.layer2": 0.2}
    """
    custom_lr_parameters = dict(
        (group_name, {"params": [], "lr": learning_rate * lr_factor})
        for (group_name, lr_factor) in lr_multipliers.items()
    )
    custom_lr_parameters["default"] = {"params": [], "lr": learning_rate}

    for parameter_name, parameter in model.named_parameters():
        if not parameter.requires_grad:
            continue

        matches = False
        for group_name, lr in lr_multipliers.items():
            if str.startswith(parameter_name, group_name):
                custom_lr_parameters[group_name]["params"].append(parameter)
                matches = True
                break

        if not matches:
            custom_lr_parameters["default"]["params"].append(parameter)

    return list(custom_lr_parameters.values())


def get_optimizable_parameters(model: nn.Module) -> Iterator[nn.Parameter]:
    """
    Return list of parameters with requires_grad=True from the model.
    This function allows easily get all parameters that should be optimized.
    :param model: An instance of nn.Module.
    :return: Parameters with requires_grad=True.
    """
    return filter(lambda x: x.requires_grad, model.parameters())


def freeze_model(
    module: nn.Module, freeze_parameters: Optional[bool] = True, freeze_bn: Optional[bool] = True
) -> nn.Module:
    """
    Change 'requires_grad' value for module and it's child modules and
    optionally freeze batchnorm modules.
    :param module: Module to change
    :param freeze_parameters: True to freeze parameters; False - to enable parameters optimization.
        If None - current state is not changed.
    :param freeze_bn: True to freeze batch norm; False - to enable BatchNorm updates.
        If None - current state is not changed.
    :return: None
    """
    bn_types = nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.SyncBatchNorm

    if freeze_parameters is not None:
        for param in module.parameters():
            param.requires_grad = not freeze_parameters

    if freeze_bn is not None:
        if isinstance(module, bn_types):
            module.track_running_stats = not freeze_bn

        for m in module.modules():
            if isinstance(m, bn_types):
                module.track_running_stats = not freeze_bn

    return module



from __future__ import absolute_import

from .common import *
from .deeplab import *
from .fpn_cat import *
from .fpn_sum import *
from .hrnet import *
from .pyramid_pooling import *
from .unet import *
from .unet_v2 import *
from .upernet import *
from .can import *
from .bifpn import *



"""Filesystem utilities

"""

import glob
import os
from pathlib import Path
from typing import Union, List

import cv2
import numpy as np

__all__ = [
    "auto_file",
    "change_extension",
    "find_images_in_dir",
    "find_images_in_dir_recursive",
    "find_in_dir",
    "find_in_dir_glob",
    "find_in_dir_with_ext",
    "find_subdirectories_in_dir",
    "has_ext",
    "has_image_ext",
    "id_from_fname",
    "read_image_as_is",
    "read_rgb_image",
]

COMMON_IMAGE_EXTENSIONS = [".bmp", ".png", ".jpeg", ".jpg", ".tiff", ".tif"]


def has_ext(fname: str, extensions: Union[str, List[str]]) -> bool:
    if not isinstance(extensions, (str, list)):
        raise ValueError("Argument extensions must be either string or list of strings")
    if isinstance(extensions, str):
        extensions = [extensions]
    extensions = set(map(str.lower, extensions))

    name, ext = os.path.splitext(fname)
    return ext.lower() in extensions


def has_image_ext(fname: str) -> bool:
    return has_ext(fname, COMMON_IMAGE_EXTENSIONS)


def find_in_dir(dirname: str) -> List[str]:
    return [os.path.join(dirname, fname) for fname in sorted(os.listdir(dirname))]


def find_subdirectories_in_dir(dirname: str) -> List[str]:
    """
    Retrieve list of subdirectories (non-recursive) in the given directory.
    Args:
        dirname: Target directory name

    Returns:
        Sorted list of absolute paths to directories
    """
    all_entries = find_in_dir(dirname)
    return [entry for entry in all_entries if os.path.isdir(entry)]


def find_in_dir_with_ext(dirname: str, extensions: Union[str, List[str]]) -> List[str]:
    return [os.path.join(dirname, fname) for fname in sorted(os.listdir(dirname)) if has_ext(fname, extensions)]


def find_images_in_dir(dirname: str) -> List[str]:
    return [fname for fname in find_in_dir(dirname) if has_image_ext(fname) and os.path.isfile(fname)]


def find_images_in_dir_recursive(dirname: str) -> List[str]:
    return [
        fname
        for fname in glob.glob(os.path.join(dirname, "**"), recursive=True)
        if has_image_ext(fname) and os.path.isfile(fname)
    ]


def find_in_dir_glob(dirname: str, recursive=False):
    files = list(glob.iglob(dirname, recursive=recursive))
    return list(sorted(files))


def id_from_fname(fname: str) -> str:
    return os.path.splitext(os.path.basename(fname))[0]


def change_extension(fname: Union[str, Path], new_ext: str) -> Union[str, Path]:
    if isinstance(fname, str):
        return os.path.splitext(fname)[0] + new_ext
    elif isinstance(fname, Path):
        if new_ext[0] != ".":
            new_ext = "." + new_ext
        return fname.with_suffix(new_ext)
    else:
        raise RuntimeError(
            f"Received input argument `fname` for unsupported type {type(fname)}. Argument must be string or Path."
        )


def auto_file(filename: str, where: str = ".") -> str:
    """Get a full path to file using it's name.
    This function recisively search for matching filename in @where and returns single match.
    :param where:
    :param filename:
    :return:
    """
    if os.path.isabs(filename):
        return filename

    prob = os.path.join(where, filename)
    if os.path.exists(prob) and os.path.isfile(prob):
        return prob

    files = list(glob.iglob(os.path.join(where, "**", filename), recursive=True))
    if len(files) == 0:
        raise FileNotFoundError("Given file could not be found with recursive search:" + filename)

    if len(files) > 1:
        raise FileNotFoundError(
            "More than one file matches given filename. Please specify it explicitly:\n" + "\n".join(files)
        )

    return files[0]


def read_rgb_image(fname: Union[str, Path]) -> np.ndarray:
    """
    Read RGB image from filesystem in RGB color order.
    Note: By default, OpenCV loads images in BGR memory order format.
    :param fname: Image file path
    :return: A numpy array with a loaded image in RGB format
    """
    if type(fname) != str:
        fname = str(fname)

    image = cv2.imread(fname, cv2.IMREAD_COLOR)
    if image is None:
        raise IOError(f'Cannot read image "{fname}"')

    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB, dst=image)
    return image


def read_image_as_is(fname: Union[str, Path]) -> np.ndarray:
    if type(fname) != str:
        fname = str(fname)
    image = cv2.imread(fname, cv2.IMREAD_UNCHANGED)
    if image is None:
        raise IOError(f'Cannot read image "{fname}"')
    return image



from collections import OrderedDict
from functools import partial
from typing import Union, List, Dict

from pytorch_toolbelt.modules import conv1x1, UnetBlock, ACT_RELU, ABN, ACT_SWISH
from pytorch_toolbelt.modules import encoders as E, DepthToSpaceUpsample2d
from pytorch_toolbelt.modules.decoders import UNetDecoder
from pytorch_toolbelt.modules.encoders import EncoderModule
from torch import nn, Tensor
from torch.nn import functional as F


__all__ = ["UnetSegmentationModel", "resnet34_unet32_s2", "resnet34_unet64_s4", "hrnet34_unet64"]


class UnetSegmentationModel(nn.Module):
    def __init__(
        self,
        encoder: EncoderModule,
        unet_channels: Union[int, List[int]],
        num_classes: int = 1,
        dropout=0.25,
        full_size_mask=True,
        activation=ACT_RELU,
        upsample_block=nn.UpsamplingNearest2d,
        last_upsample_block=None,
    ):
        super().__init__()
        self.encoder = encoder

        abn_block = partial(ABN, activation=activation)
        self.decoder = UNetDecoder(
            feature_maps=encoder.channels,
            decoder_features=unet_channels,
            unet_block=partial(UnetBlock, abn_block=abn_block),
            upsample_block=upsample_block,
        )

        if last_upsample_block is not None:
            self.mask = nn.Sequential(
                OrderedDict(
                    [("drop", nn.Dropout2d(dropout)), ("conv", last_upsample_block(unet_channels[0], num_classes))]
                )
            )
        else:
            self.last_upsample_block = None
            self.mask = nn.Sequential(
                OrderedDict([("drop", nn.Dropout2d(dropout)), ("conv", conv1x1(unet_channels[0], num_classes))])
            )

        self.full_size_mask = full_size_mask

    def forward(self, x: Tensor) -> Dict[str, Tensor]:
        x_size = x.size()
        x = self.encoder(x)
        x = self.decoder(x)

        # Decode mask
        mask = self.mask(x[0])
        if self.full_size_mask and mask.size()[2:] != x_size:
            mask = F.interpolate(mask, size=x_size[2:], mode="bilinear", align_corners=False)

        return mask


def resnet34_unet32_s2(input_channels=3, num_classes=1, dropout=0.2, pretrained=True):
    encoder = E.Resnet34Encoder(pretrained=pretrained, layers=[0, 1, 2, 3, 4])
    if input_channels != 3:
        encoder.change_input_channels(input_channels)
    return UnetSegmentationModel(
        encoder,
        num_classes=num_classes,
        unet_channels=[32, 64, 128, 256],
        activation=ACT_SWISH,
        dropout=dropout,
    )


def resnet34_unet64_s4(input_channels=3, num_classes=1, dropout=0.2, pretrained=True):
    encoder = E.Resnet34Encoder(pretrained=pretrained, layers=[1, 2, 3, 4])
    if input_channels != 3:
        encoder.change_input_channels(input_channels)
    return UnetSegmentationModel(
        encoder,
        num_classes=num_classes,
        unet_channels=[64, 128, 256],
        activation=ACT_SWISH,
        dropout=dropout,
    )


def hrnet34_unet64(input_channels=3, num_classes=1, dropout=0.2, pretrained=True):
    encoder = E.HRNetV2Encoder34(pretrained=pretrained, layers=[1, 2, 3, 4])
    if input_channels != 3:
        encoder.change_input_channels(input_channels)

    return UnetSegmentationModel(encoder, num_classes=num_classes, unet_channels=[64, 128, 256], dropout=dropout)



import torch

from ..common import EncoderModule, make_n_channel_input

__all__ = [
    "DPN68BEncoder",
    "DPN68Encoder",
    "DPN92Encoder",
    "DPN107Encoder",
    "DPN131Encoder",
]


class DPN68Encoder(EncoderModule):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import dpn

        if layers is None:
            layers = [0, 1, 2, 3]

        encoder = dpn.dpn68(pretrained=pretrained, features_only=True, out_indices=(1, 2, 3, 4))
        super().__init__([144, 320, 704, 832], [4, 8, 16, 32], layers)
        self.encoder = encoder

    def forward(self, x):
        y = self.encoder.forward(x)
        return y

    @torch.jit.unused
    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.features_conv1_1.conv = make_n_channel_input(
            self.encoder.features_conv1_1.conv, input_channels, mode, **kwargs
        )
        return self


class DPN68BEncoder(EncoderModule):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import dpn

        if layers is None:
            layers = [0, 1, 2, 3]

        encoder = dpn.dpn68b(pretrained=pretrained, features_only=True, out_indices=(1, 2, 3, 4))
        super().__init__([144, 320, 704, 832], [4, 8, 16, 32], layers)
        self.encoder = encoder

    def forward(self, x):
        y = self.encoder.forward(x)
        return y

    @torch.jit.unused
    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.features_conv1_1.conv = make_n_channel_input(
            self.encoder.features_conv1_1.conv, input_channels, mode, **kwargs
        )
        return self


class DPN92Encoder(EncoderModule):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import dpn

        if layers is None:
            layers = [0, 1, 2, 3]

        encoder = dpn.dpn92(pretrained=pretrained, features_only=True, out_indices=(1, 2, 3, 4))
        super().__init__([336, 704, 1552, 2688], [4, 8, 16, 32], layers)
        self.encoder = encoder

    def forward(self, x):
        y = self.encoder.forward(x)
        return y

    @torch.jit.unused
    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.features_conv1_1.conv = make_n_channel_input(
            self.encoder.features_conv1_1.conv, input_channels, mode, **kwargs
        )
        return self


class DPN107Encoder(EncoderModule):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import dpn

        if layers is None:
            layers = [0, 1, 2, 3]

        encoder = dpn.dpn107(pretrained=pretrained, features_only=True, out_indices=(1, 2, 3, 4))
        super().__init__([376, 1152, 2432, 2688], [4, 8, 16, 32], layers)
        self.encoder = encoder

    def forward(self, x):
        y = self.encoder.forward(x)
        return y

    @torch.jit.unused
    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.features_conv1_1.conv = make_n_channel_input(
            self.encoder.features_conv1_1.conv, input_channels, mode, **kwargs
        )
        return self


class DPN131Encoder(EncoderModule):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import dpn

        if layers is None:
            layers = [0, 1, 2, 3]
        encoder = dpn.dpn131(pretrained=pretrained, features_only=True, out_indices=(1, 2, 3, 4))
        super().__init__([352, 832, 1984, 2688], [4, 8, 16, 32], layers)
        self.encoder = encoder

    def forward(self, x):
        y = self.encoder.forward(x)
        return y

    @torch.jit.unused
    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.features_conv1_1.conv = make_n_channel_input(
            self.encoder.features_conv1_1.conv, input_channels, mode, **kwargs
        )
        return self



from typing import Dict

import torch
from catalyst.dl import (
    CriterionCallback,
    IRunner,
)
from torch import nn, Tensor

__all__ = [
    "TrainOnlyCriterionCallback",
    "PassthroughCriterionCallback",
    "LossModule",
    "LossWrapper",
]


class TrainOnlyCriterionCallback(CriterionCallback):
    """
    Computes loss only on training stage
    """

    def _compute_loss_value(self, state: IRunner, criterion):
        predictions = self._get_output(state.output, self.output_key)
        targets = self._get_input(state.input, self.input_key)

        if state.loader_name != "train":
            return torch.tensor(0, device=predictions.device, dtype=predictions.dtype)

        loss = criterion(predictions, targets)
        return loss

    def _compute_loss_key_value(self, state: IRunner, criterion):
        output = self._get_output(state.output, self.output_key)
        input = self._get_input(state.input, self.input_key)

        if state.loader_name != "train":
            return torch.tensor(0, device=output.device, dtype=output.dtype)

        loss = criterion(**output, **input)
        return loss


class PassthroughCriterionCallback(CriterionCallback):
    """
    Returns one of model's outputs as loss values
    """

    def __init__(self, output_key, multiplier=1.0):
        super().__init__(output_key=output_key, prefix=output_key, multiplier=multiplier)

    def _compute_loss_value(self, state: IRunner, criterion):
        loss = self._get_output(state.output, self.output_key)
        return loss.mean()

    def _compute_loss_key_value(self, state: IRunner, criterion):
        loss = self._get_output(state.output, self.output_key)
        return loss.mean()


class LossModule(nn.Module):
    def __init__(self, output_key: str, target_key: str, loss_fn):
        super().__init__()
        self.output_key = output_key
        self.target_ley = target_key
        self.loss_fn = loss_fn

    def forward(self, outputs, targets):  # skipcq: PYL-W0221
        return self.loss_fn(outputs[self.output_key], targets[self.target_ley])


class LossWrapper(nn.Module):
    """
    A wrapper module around model that computes one or many loss functions and extends output dictionary with
    their values. The point of this wrapper is that loss computed on each GPU node in parallel.

    Usage:
    >>> from catalyst.dl import SupervisedRunner
    >>> runner = SupervisedRunner(input_key=None, output_key=None, device="cuda")
    >>> runner._default_experiment = ParallelLossSupervisedExperiment
    >>> loss_modules = {
    >>>     "my_loss": LossModule(
    >>>         output_key="logits",
    >>>         target_key="targets",
    >>>         loss_fn=nn.BCEWithLogitsLoss(),
    >>>     )}
    >>> loss_callback = PassthroughCriterionCallback("my_loss")
    >>> runner.train(
    >>>     callbacks=[loss_callback, ...]
    >>>     model=LossWrapper(model, "image", loss_modules),
    >>>     ...)

    Note, that SupervisedRunner adds default CriterionCallback
    """

    def __init__(self, model: nn.Module, input_key: str, losses: Dict[str, LossModule]):
        super().__init__()
        self.model = model
        self.input_key = input_key
        self.loss_names = list(losses.keys())
        self.losses = nn.ModuleList([losses[key] for key in self.loss_names])

    def forward(self, **input: Dict[str, Tensor]) -> Dict[str, Tensor]:  # skipcq: PYL-W0221
        output: Dict[str, Tensor] = self.model(input[self.input_key])

        for output_loss_key, loss in zip(self.loss_names, self.losses):
            output[output_loss_key] = loss(output, input)

        return output

    def state_dict(self, destination=None, prefix="", keep_vars=False):
        return self.model.state_dict()



import torch
from torch import nn

__all__ = ["Normalize"]


class Normalize(nn.Module):
    def __init__(self, mean, std):
        super().__init__()
        self.register_buffer("mean", torch.tensor(mean).float().reshape(1, len(mean), 1, 1).contiguous())
        self.register_buffer("std", torch.tensor(std).float().reshape(1, len(std), 1, 1).reciprocal().contiguous())

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return (input.type_as(self.mean) - self.mean) * self.std
